"""Convert MusicXML to PDF format via LilyPond."""

import shutil
import subprocess
import tempfile
import webbrowser
from pathlib import Path

from amads.io.displayscore import suppress_external_open


def lilypond_path_help(
    filename: Path | str, lilypond: bool, display: bool, xml: bool
) -> tuple[Path, Path | None, Path | None]:
    """There are multiple messy ways to determine file names and paths.
    This helper function serves both music21_pdf_export and
    music21_xml_pdf_export.

    If `lilypond`, then ly_path will be filename; otherwise, ly_path
    will be a temp file.

    Parameters
    ----------
    filename : Path | str
        The filename for the pdf export.
    lilypond: bool
        True to retain the lilypond (.ly) file
    display: bool
        True if the pdf file will be displayed, meaning filename is optional.
    xml: bool
        True if we also need an xml file path
    """
    ly_path: Path | None = None
    pdf_path: Path | None = None
    xml_path: Path | None = None

    work_dir = None
    stem = "score"

    if not filename and not lilypond and not display:
        raise ValueError(
            "filename must be provided if lilypond and" " display are False"
        )
    elif filename:
        ext = Path(filename).suffix.lower()
        if lilypond:
            ly_path = Path(filename)
            if ext != ".ly":
                raise ValueError(
                    f"filename extension {ext} must be .ly when"
                    " lilypond is True"
                )
        else:
            pdf_path = Path(filename)
            if ext != ".pdf":
                raise ValueError(
                    f"filename extension {ext} must be .pdf when"
                    " lilypond is False"
                )

    if not ly_path:
        work_dir = Path(tempfile.mkdtemp(prefix="amads_lilypond_"))
        ly_path = work_dir / f"{stem}.ly"
    if not pdf_path and not lilypond:
        if not work_dir:
            work_dir = Path(tempfile.mkdtemp(prefix="amads_lilypond_"))
        pdf_path = work_dir / f"{stem}.pdf"
    if xml:
        if not work_dir:
            work_dir = Path(tempfile.mkdtemp(prefix="amads_lilypond_"))
        xml_path = work_dir / f"{stem}.musicxml"

    return (ly_path, pdf_path, xml_path)


def lilypond_to_pdf(
    ly_path: Path | str, pdf_path: Path | str, display: bool = False
) -> None:
    """Convert a Lilypond (.ly) file to PDF.

    Parameters
    ----------
    ly_path : Path | str
        The path to the LilyPond file.
    pdf_path : Path | str
        The path to the PDF file.
    display : bool, optional
        If True, open the generated PDF in the default viewer.

    Raises
    ------
    RuntimeError
        If LilyPond fails to convert the file.
    """

    # we have pdf_path ending in ".pdf" but lilypond will add ".pdf", so
    # remove the extension. Then run LilyPond to create the PDF.
    if not isinstance(ly_path, (str, Path)):
        raise ValueError(f"ly_path must be a string or Path, got {ly_path}.")
    if not isinstance(pdf_path, (str, Path)):
        raise ValueError(f"pdf_path must be a string or Path, got {pdf_path}.")
    pdf_path_obj = Path(pdf_path)
    output_base = pdf_path_obj.with_suffix("")
    command = ["lilypond", "-o", str(output_base), str(ly_path)]
    result = subprocess.run(command, capture_output=True, text=True)
    if result.returncode != 0:
        raise RuntimeError(
            "LilyPond failed. If `lilypond` did not even start, ensure that\n"
            "`lilypond` is installed and on PATH.\n"
            "In VS Code: You may need to set the PATH in your settings.json,\n"
            "    although there are at least four separate settings related\n"
            "    to setting the PATH! A more direct approach is hard-code\n"
            "    PATH in .vscode/launch.json for the configuration(s) you are\n"
            '    using: "env": {"PATH": "path-to-lilypond:${env:PATH}"}\n'
            f"Command: {' '.join(command)}\n"
            f"stdout:\n{result.stdout}\n"
            f"stderr:\n{result.stderr}"
        )

    if display:
        if suppress_external_open():
            print(f"PDF display suppressed during tests, wrote {pdf_path}.")
        else:
            webbrowser.open(pdf_path_obj.resolve().as_uri())  # type: ignore


def _add_measure_numbers_to_xml(
    xml_content: str, xml_path: Path | str, xml_temp_file: bool
) -> Path:
    """Add measure numbers to a MusicXML file if they are missing, because
    musicxml2ly requires measure numbers.

    Parameters
    ----------
    xml_content : str
        The content of the MusicXML file.
    xml_path : Path | str
        The path to the MusicXML file.
    xml_temp_file : bool
        If True, xml_path is a temporary file that can be overwritten with
        the modified content.

    Returns
    -------
    Path
        The path to the modified MusicXML file.
    """
    if not xml_temp_file:
        work_dir = Path(tempfile.mkdtemp(prefix="amads_lilypond_"))
        new_xml_path = work_dir / "score_with_measure_numbers.musicxml"
    else:
        new_xml_path = Path(xml_path)
    with open(new_xml_path, "w", encoding="utf-8") as f:
        measure_count = 0
        pos = 0
        next_part_loc = xml_content.find("<part", pos)
        if next_part_loc == -1:
            raise RuntimeError(f"No <part> tags found in {xml_path}.")
        while True:
            loc = xml_content.find("<measure", pos)
            if loc == -1:
                f.write(xml_content[pos:])
                break
            loc2 = xml_content.find(">", loc)
            if loc2 == -1:
                raise RuntimeError(
                    f"No closing '>' found in {xml_path} after <measure."
                )
            measure_tag = xml_content[loc : loc2 + 1]
            if loc > next_part_loc:
                measure_count = 0
                next_part_loc = xml_content.find("<part", loc)
                if next_part_loc == -1:
                    next_part_loc = len(xml_content) + 1
            if "number=" not in measure_tag:
                measure_count += 1
                new_measure_tag = (
                    measure_tag[:-1] + f' number="{measure_count}">'
                )
                f.write(xml_content[pos:loc] + new_measure_tag)
            else:
                f.write(xml_content[pos : loc2 + 1])
            pos = loc2 + 1
    return new_xml_path


def musicxml_to_lilypond(
    xml_path: Path | str, ly_path: Path | str, xml_temp_file=False
) -> None:
    """Convert a MusicXML file to a LilyPond file.

    Parameters
    ----------
    xml_path : Path | str
        The path to the MusicXML file.
    ly_path : Path | str
        The path to the LilyPond file.

    Raises
    ------
    RuntimeError
        If musicxml2ly fails to convert the file.
    """
    musicxml2ly = shutil.which("musicxml2ly")
    if not musicxml2ly:
        raise RuntimeError(
            "Could not find musicxml2ly. Ensure LilyPond is installed and "
            "musicxml2ly is on PATH.\n"
            "In VS Code: You may need to set the PATH in your settings.json,\n"
            "    although there are at least four separate settings related\n"
            "    to setting the PATH! A more direct approach is hard-code\n"
            "    PATH in .vscode/launch.json for the configuration(s) you are\n"
            '    using: "env": {"PATH":'
            ' "path-to-musicxml2ly:${env:PATH}"}\n'
        )

    # make sure xml measures have numbers, otherwise musicxml2ly will fail.
    # Partitura apparently omits measure numbers, which is a problem.
    with open(xml_path, "r", encoding="utf-8") as f:
        xml_content = f.read()
    loc = xml_content.find("<measure")
    if loc == -1:
        raise RuntimeError(f"No <measure> tags found in {xml_path}.")
    loc2 = xml_content.find(">", loc)
    if loc2 == -1:
        raise RuntimeError(
            f"No closing '>' found in {xml_path} after <measure."
        )
    measure_tag = xml_content[loc : loc2 + 1]
    if "number=" not in measure_tag:
        xml_path = _add_measure_numbers_to_xml(
            xml_content, xml_path, xml_temp_file
        )

    # now xml_path is either the original xml_path (if it already had
    # measure numbers) or a new file was created with measure numbers.
    command = [musicxml2ly, "-o", str(ly_path), str(xml_path)]
    result = subprocess.run(command, capture_output=True, text=True)
    if result.returncode != 0:
        raise RuntimeError(
            "musicxml2ly failed. Ensure LilyPond is installed and on PATH.\n"
            f"Command: {' '.join(command)}\n"
            f"stdout:\n{result.stdout}\n"
            f"stderr:\n{result.stderr}"
        )



"""
Provides the `skyline` function
"""

from typing import List, Optional, cast

from amads.core.basics import Note, Part, Score


def skyline(score: Score, threshold: float = 0.1) -> Score:
    """
    Finds the skyline of a musical score.

    Filters a score, removing any note that is below another note.
    There are tricky edge cases:

    - A higher note can occur while a lower note is still sounding. The lower
      note is shortened to end at the onset of the upper note.
    - A higher note can follow a lower note very quickly: Rather than setting
      the lower note's duration to a very small value, the lower note is
      completely removed and ignored. The lower bound on duration is set by
      the `threshold` parameter.
    - A rolled chord with 10 notes starts at the bottom, and every 0.05
      quarter notes, a new note enters. So the previous rule applies to
      each note, making the top note is a full 0.45 quarters after the
      first one. Even in this case, the previous rule is applied repeatedly,
      leaving a gap (rest) of at least 0.45 quarters.
    - An upper note of a melody sustains in a legato fashion past the next,
      but lower, note of the melody. Although, musically, the upper note
      should be shortened and we should keep the lower note, the “skyline”
      concept says the top note has priority, so the lower note is ignored
      if the overlap is greater than `threshold`.
    - It is common to have melodies in lower voices or in MIDI arrangements
      to have very high accompaniment notes in a non-melodic track. This
      algorithm just fails to find the melody in those cases.

    Parameters
    ----------
    score : Score
        The musical score to filter
    threshold : float
        The threshold for quickly followed notes (default 0.1) and allowed
        overlap. Processing onsets in time order, if an onset is within
        `threshold` of the previous onset, the two notes are considered
        to be conconcurrent, and only the top note is considered in
        constructing the skyline. In the case of processing a note that
        is lower in pitch than the current skyline, we ignore the note
        if the skyline extends more than `threshold` beyond the note's
        onset. Otherwise, we shorten the skyline duration to end at the
        note onset and append the note to the skyline.

    Returns
    -------
    Score
        A new score containing the “skyline” notes

    Algorithm
    ---------
    The basic idea is to scan notes and copy them to skyline, a Part object
    belonging to a new Score. We can use shallow copy because notes are
    already deep-copied from score after merge_tied_notes.

    In the outer looop, we test each note to see if it is below the skyline
    as it exists so far. Since we process in order, we know each note cannot
    start before any note in the skyline. If the note is higher than the
    most recent skyline note (so far), there are two cases:

    1. The new note is approximately concurrent with the most recent skyline
       note: Replace the most recent skyline note with this one.
    2. The new note is after the most recent skyline note: Append the new
       note to the skyline. If the previous note overlaps the new note,
       adjust the previous note's duration to end at the onset of the new
       note.

    A consequence of this algorithm is that a very long low note will
    be shortened to the onset time of a new note, so a piano roll like
    this:
    ```
                               ----------
             ------------------------------------------
    ```
    will result in this:
    ```
                               ----------
             ------------------          (nothing here)
    ```
    rather than this:
    ```                        ----------
             ------------------          --------------
    ```

    Another consequence is that since skyline notes are never lengthened,
    there can be gaps in the skyline. It can look like this, where gaps
    (rests) can occur between notes:
    ```
                           ------      -----
             --------            ------         -------
    ```
    """
    # this code is based on get_sorted_notes():
    score = score.flatten(collapse=True)  # deep copies the score
    skyline: Part = cast(
        Part, score.content[0]
    )  # types: ignore (retrieves the Part)
    notes: List[Note] = cast(
        List[Note], skyline.content
    )  # (content is all Notes)
    skyline.content = []  # we will construct skyline from notes here

    prev_note: Optional[Note] = None
    for note in notes:
        if prev_note is None:
            skyline.content.append(note)
            prev_note = note

        # ignore notes that are below last note in skyline
        elif note.key_num < prev_note.key_num and (
            note.onset < prev_note.offset - threshold  # overlap
            or note.onset < prev_note.onset + threshold
        ):  # concurrent
            continue

        elif note.duration < threshold:
            continue

        # see if note is concurrent and higher
        if (
            note.onset < prev_note.onset + threshold
            and note.key_num >= prev_note.key_num
        ):
            skyline.content.pop()  # replace prev_note
            skyline.content.append(note)
            prev_note = note

        elif (
            note.onset >= prev_note.offset - threshold
            or note.key_num >= prev_note.key_num
        ):
            skyline.content.append(note)
            prev_note = note
            if prev_note.offset > note.onset:
                prev_note.offset = note.onset

        else:
            assert False, "Unexpected condition, implementation error detected"
    return score



"""
This module serves to map out metrical hierarchies in a number of different ways and
to express the relationship between
the notes in
and
the hierarchy of
a metrical cycle.

Uses include identifying notes that traverse metrical levels,
for analysis (e.g., as a cycle of syncopation)
and notation (e.g., re-notating to reflect the
within-cycle notational conventions).


<small>**Author**: Mark Gotham</small>
"""

__author__ = "Mark Gotham"

import math
from typing import Optional, Union

import numpy as np

# ------------------------------------------------------------------------------


def is_non_negative_integer_power_of_two(n: float) -> bool:
    """
    Checks if a number is a power of 2.

    Examples
    --------
    >>> is_non_negative_integer_power_of_two(0)
    False

    >>> is_non_negative_integer_power_of_two(0.5)
    False

    >>> is_non_negative_integer_power_of_two(1)
    True

    >>> is_non_negative_integer_power_of_two(2)
    True

    >>> is_non_negative_integer_power_of_two(3)
    False

    >>> is_non_negative_integer_power_of_two(4)
    True
    """
    if n <= 0:  # also catches type error if non-numeric
        return False
    if not isinstance(n, int):
        if int(n) == n:
            n = int(n)
        else:
            return False
    return n > 0 and (n & (n - 1)) == 0


def switch_pulse_length_beat_type(
    pulse_length_or_beat_type: Union[float, np.ndarray]
):
    """
    Switch between a pulse length and beat type.
    Accepts numeric values or numpy arrays thereof.
    Note that a float of vale 0 will raise a
    `ZeroDivisionError: division by zero`,
    but a numpy array will map any 0s to `inf` without error.

    Examples
    --------
    >>> switch_pulse_length_beat_type(0.5)
    8.0

    >>> switch_pulse_length_beat_type(8)
    0.5

    >>> switch_pulse_length_beat_type(np.array([0.5, 8]))
    array([8. , 0.5])
    """
    return 4 / pulse_length_or_beat_type


class StartTimeHierarchy:
    """
    Encoding metrical structure as a hierarchy of start times.

    A representation of metrical levels in terms of starts expressed by quarter length
    from the start of the cycle.

    Parameters
    ----------
    start_hierarchy
        Users can specify the `start_hierarchy` directly and completely from scratch.
        Use this for advanced, non-standard metrical structures
        including those without 2-/3- grouping, or even nested hierarchies,
        as well as for (optionally) encoding micro-timing directly into the
        metrical structure.
        The only “well-formed” criteria we expect are
        use of 0.0 and full cycle length at the top level, and
        presence of all timepoints from one level in each subsequent level.
        For creating this information from pulse lengths, time signatures,
        and more see the `to_start_hierarchy` methods on those classes.
    names
        Optionally create a dict mapping temporal positions to names.
        Currently, this supports one textual value per temporal position (key),
        e.g., {0.0: "ta", 1.0: "ka", 2.0: "di", 3.0: "mi"}.

    """

    def __init__(
        self,
        start_hierarchy: list[list],
        names: Optional[dict] = None,
    ):
        self.start_hierarchy = start_hierarchy
        self.cycle_length = self.start_hierarchy[0][-1]
        self.pulse_lengths = None

        if names:
            for key in names:
                assert isinstance(key, float)
                assert isinstance(names[key], str)
        self.names = names

    def coincident_pulse_list(
        self,
        granular_pulse: float,
    ) -> list:
        """
        Create a flat list setting out the
        number of intersecting pulses at each successive position in a metrical cycle.

        For example,
        the output [4, 1, 2, 1, 3, 1, 2, 1]
        refers to a base pulse unit of 1,
        with addition pulse streams accenting every 2nd, 4th, and 8th position.


        Parameters
        --------
        granular_pulse
            The pulse value of the fastest level to consider e.g., 1, or 0.25.

        Examples
        --------

        You can currently set the `granular_pulse` value to anything (this may change).
        For instance, in the pair of examples below,
        first we have a `granular_pulse` that's present in the input,
        and then a case using a faster level that's not present (this simply pads the data out):

        >>> hierarchy = StartTimeHierarchy([[0.0, 4.0], [0.0, 2.0, 4.0], [0.0, 1.0, 2.0, 3.0, 4.0]])
        >>> hierarchy.coincident_pulse_list(granular_pulse=1)
        [3, 1, 2, 1]

        Now, changing the `granular_pulse` for a bit of over-sampling:

        >>> hierarchy.coincident_pulse_list(granular_pulse=0.5)
        [3, 0, 1, 0, 2, 0, 1, 0]

        """
        cycle_length = self.start_hierarchy[0][-1]

        for level in self.start_hierarchy:
            assert level[-1] == cycle_length

        steps = int(cycle_length / granular_pulse)
        granular_level = [granular_pulse * count for count in range(steps)]

        def count_instances(nested_list, target):
            return sum([sublist.count(target) for sublist in nested_list])

        coincidences = []
        for target in granular_level:
            coincidences.append(count_instances(self.start_hierarchy, target))

        return coincidences

    def to_pulse_lengths(self):
        """
        Check if levels have a regular pulse and if so, return the pulse length value.

        Returns
        -------
        list
            Returns a list of pulse values corresponding to the start hierarchy, of the same length.
            If a level is not regular, the list is populated with None.

        Examples
        --------

        >>> hierarchy = StartTimeHierarchy([[0.0, 4.0], [0.0, 2.0, 4.0], [0.0, 1.0, 2.0, 3.0, 4.0]])
        >>> hierarchy.to_pulse_lengths()
        >>> hierarchy.pulse_lengths
        [4.0, 2.0, 1.0]

        >>> uneven = StartTimeHierarchy([[0.0, 4.0], [0.0, 3.0, 4.0], [0.0, 1.0, 2.0, 3.0, 4.0]])
        >>> uneven.to_pulse_lengths()
        >>> uneven.pulse_lengths
        [4.0, None, 1.0]

        """

        def test_one(level: list):
            diffs = set(
                [level[i + 1] - level[i] for i in range(len(level) - 1)]
            )
            if len(diffs) > 1:
                return None
            return float(list(diffs)[0])

        self.pulse_lengths = [test_one(level) for level in self.start_hierarchy]

    def add_faster_levels(self, minimum_beat_type: int = 64):
        """
        Recursively add faster levels until the `minimum_beat_type` value
        The `minimum_beat_type` is subject to the same constraints as the `beat_types` ("denominators")
        i.e., powers of 2 (1, 2, 4, 8, 16, 32, 64, ...).
        The default = 64 for 64th note.

        Parameters
        ----------
        minimum_beat_type
            Recursively create further levels down to this value.
            Must be power of two.
            Defaults to 64 for 64th notes.

        Raises
        ------
        ValueError
            if the currently fastest level of a `starts_hierarchy` is not periodic,
            or if either of the fastest level or `minimum_beat_type` are not powers
            of 2. Set the `starts_hierarchy` manually in these non-standard cases.

        Examples
        --------
        >>> hierarchy = StartTimeHierarchy([[0.0, 4.0], [0.0, 2.0, 4.0]])
        >>> hierarchy.start_hierarchy
        [[0.0, 4.0], [0.0, 2.0, 4.0]]

        >>> hierarchy.to_pulse_lengths()
        >>> hierarchy.pulse_lengths
        [4.0, 2.0]

        >>> hierarchy.add_faster_levels(minimum_beat_type=4)
        >>> hierarchy.start_hierarchy
        [[0.0, 4.0], [0.0, 2.0, 4.0], [0.0, 1.0, 2.0, 3.0, 4.0]]

        >>> hierarchy.pulse_lengths
        [4.0, 2.0, 1.0]

        >>> hierarchy.add_faster_levels(minimum_beat_type=8)
        >>> hierarchy.start_hierarchy[-1]
        [0.0, 0.5, 1.0, 1.5, 2.0, 2.5, 3.0, 3.5, 4.0]

        >>> len(hierarchy.start_hierarchy)
        4

        >>> hierarchy.pulse_lengths
        [4.0, 2.0, 1.0, 0.5]

        """
        self.to_pulse_lengths()
        assert self.pulse_lengths is not None, ""
        fastest = self.pulse_lengths[-1]
        if fastest is None:
            raise ValueError(
                "Fastest level is not regular. Use case unsupported."
            )
        if not is_non_negative_integer_power_of_two(
            switch_pulse_length_beat_type(
                fastest
            )  # from pulse length to beat type
        ):
            raise ValueError(
                f"Fastest level ({fastest}) is not a power of 2. Use case unsupported."
            )
        if not is_non_negative_integer_power_of_two(minimum_beat_type):
            raise ValueError(
                f"The `minimum_beat_type` ({minimum_beat_type}) is not a power of 2. Use case unsupported."
            )

        fastest_beat_type = switch_pulse_length_beat_type(fastest)  # TODO
        fastest_beat_type_exponent = int(math.log2(fastest_beat_type))
        minimum_beat_type_exponent = int(math.log2(minimum_beat_type))

        new_beat_types = [
            2**x
            for x in range(
                fastest_beat_type_exponent + 1, minimum_beat_type_exponent + 1
            )
        ]
        new_pulses = [
            switch_pulse_length_beat_type(beat_type)
            for beat_type in new_beat_types
        ]
        self.pulse_lengths += new_pulses
        self.pulse_lengths = [x for x in self.pulse_lengths if x is not None]
        self.pulse_lengths = sorted(
            list(set(self.pulse_lengths)), key=abs, reverse=True
        )
        fake_meter = PulseLengths(
            pulse_lengths=new_pulses, cycle_length=self.cycle_length
        )
        self.start_hierarchy += fake_meter.to_start_hierarchy()


# ------------------------------------------------------------------------------


class TimeSignature:
    """
    Represent the _notational_ time signature object.
    # TODO consider aligning and merging with basics.TimeSignature, this PR shows some of how that would work.

    Parameters
    ----------
    beats
        The "numerator" of the time signature: beats per cycle, a number (int or fraction) or a lists thereof.
    beat_type
        the so-called "denominator" of the time signature: a whole number power of 2
        (1, 2, 4, 8, 16, 32, 64, ...).
        No so-called "irrational" meters yet (e.g., 2/3), sorry!
    as_string
        An alternative way of creating this object from a string representation.
        See notes at `TimeSignature.from_string`.

    """

    def __init__(
        self,
        beats: Optional[tuple[int]] = None,
        beat_type: Optional[int] = None,
        # delta: Optional[float] = 0,  # TODO if merging with basics
        as_string: Optional[str] = None,
    ):
        self.beats = beats
        self.one_beat_value = None
        self.beat_type = beat_type
        self.as_string = as_string
        if (self.beats is None) and (self.beat_type is None):
            self.from_string()
        self.check_valid()

        self.cycle_length = sum(self.beats) * 4 / self.beat_type
        self.pulses = None
        self.get_pulses()

    def from_string(self):
        """
        Given a signature string, extract the constituent parts and create an object.
        The string must take the form `<beat>/<beat_type>`
        with exactly one “/” separating the two (spaces are ignored).
        The string does not change.

        The `<beat>` (“numerator”) part may be a number (including 5 and 7 which
        are supported) or more than one number separated by the “+” symbol.
        For example, when encoding “5/4”, use the total value only to avoid
        segmentation above the denominator level (“5/4”)
        or the X+Y form to explicitly distinguish between “2+3” and “3+2”.
        I.e., “5/” time signatures have no 3+2 or 2+3 division by default.
        See examples on `TimeSignature.to_starts_hierarchy`.

        Finally, although we support and provide defaults for time signatures
        in the form “2+3/8”, there is no such support for more than one “/”
        (i.e., the user must build cases like “4/4 + 3/8” explicitly
        according to how they see it).

        Examples
        --------

        >>> ts_4_4 = TimeSignature(as_string="4/4")
        >>> ts_4_4.beats # Tuple of one element
        (4,)

        >>> ts_4_4.beat_type
        4

        """
        beats, beat_type = self.as_string.split("/")

        self.beats = tuple([int(x) for x in beats.split("+")])
        self.beat_type = int(beat_type)

    def check_valid(self):
        """
        Check the validity of the input.

         - `.beats` must be an integer or a list/tuple thereof.
         - `.beat_type` must be a single integer power of two.
        """
        # beats  # TODO this check may be overdoing it
        if self.beats:
            assert isinstance(self.beats, tuple)
            for b in self.beats:
                assert isinstance(b, int)

        # beat_type  # TODO this is the part we want to actively check
        if not is_non_negative_integer_power_of_two(self.beat_type):
            raise ValueError(
                f"Beat type set as {self.beat_type} is invalid: must be a non-negative integer power of 2."
            )

    def get_pulses(self):
        """
        Create an unordered set for the regular pulses present in this time signature.

        This will include the full cycle and beat type (“denominator”) value,
        e.g., in “3/4” the pulse lengths are 3.0 (full cycle) and 1.0 (beat type).
        If there are other regular levels between the two, they will be added
        only if the user has first called `fill_2s_3s` (it does not run by default).
        For instance, the splitting of 4 into 2+2 is user choice (see `fill_2s_3s`)
        With this split, this “4/4” has pulse lengths of 4.0 (full cycle)
        and 1.0 (beat type) as well as 2.0 given that the two twos are of one kind.
        In “2+3/4” there is no such 2.0 (or 3.0) regularity, and so no pulse is
        created for that level.

        Examples
        --------
        >>> ts_4_4 = TimeSignature(as_string="4/4")
        >>> ts_4_4.pulses
        [4.0, 1.0]

        >>> ts_4_4.fill_2s_3s()
        >>> ts_4_4.pulses
        [4.0, 2.0, 1.0]

        >>> ts_6_8 = TimeSignature(as_string="6/8")
        >>> ts_6_8.pulses
        [3.0, 0.5]

        >>> ts_6_8.fill_2s_3s()
        >>> ts_6_8.pulses
        [3.0, 1.5, 0.5]

        """
        pulses = [float(self.cycle_length), 4 / self.beat_type]

        first_beat_to_pulse = self.beats[0] * 4 / self.beat_type

        if len(self.beats) == 1:  # one beat type
            pulses.append(first_beat_to_pulse)
            self.one_beat_value = self.beats[0]
        elif len(self.beats) > 1:  # 2+ beats
            if (
                len(set(self.beats)) == 1
            ):  # duplicate of the same e.g., (2, 2), so still one consistent pulse.
                self.one_beat_value = self.beats[0]
                pulses.append(first_beat_to_pulse)

        self.pulses = sorted(list(set(pulses)), key=abs, reverse=True)

    def fill_2s_3s(self):
        """
        Optionally, add pulse values to follow the conventions of the time signatures.

        Enforcing 2- and 3-grouping, this only applies to cases with a
        single beat in the "numerator". For instance,
        given a “4/4” signature, this method will add the half-cycle (pulse value 2.0),
        given a “6/8”, it will again add the half-cycle (pulse value 1.5),
        and given a “12/8”, it will add both the half- and quarter-cycle
        (pulse values 3.0 and 1.5),

        This functionality is factored out and does not run by default.
        Even if this runs, the original time signature string is unchanged,
        as is the `beats` attribute.

        Examples
        --------
        >>> ts_4_4 = TimeSignature(as_string="4/4")
        >>> ts_4_4.pulses
        [4.0, 1.0]

        >>> ts_4_4.fill_2s_3s()
        >>> ts_4_4.pulses
        [4.0, 2.0, 1.0]

        >>> ts_6_8 = TimeSignature(as_string="6/8")
        >>> ts_6_8.pulses
        [3.0, 0.5]

        >>> ts_6_8.fill_2s_3s()
        >>> ts_6_8.pulses
        [3.0, 1.5, 0.5]

        """
        metrical_mappings = {4: [2], 6: [3], 9: [3], 12: [6, 3]}

        if self.one_beat_value is not None:
            if self.one_beat_value in metrical_mappings:
                self.pulses += [
                    x * 4 / self.beat_type
                    for x in metrical_mappings[self.one_beat_value]
                ]
        self.pulses = sorted(list(set(self.pulses)), key=abs, reverse=True)

    def to_start_hierarchy(self) -> list:
        """
        Create a start hierarchy for almost any time signature

        (with constraints as noted in the top level class description
        and in the `.from_string` method).
        See below for several examples of how this handles
        specific time signatures and related assumptions,
        and note the effect of running `fill_2s_3s()`.

        Returns
        -------
        list
            Returns a list of lists with start positions by level.

        Examples
        --------

        >>> ts_4_4 = TimeSignature(as_string="4/4")
        >>> ts_4_4.pulses
        [4.0, 1.0]

        >>> test_1 = ts_4_4.to_start_hierarchy()
        >>> test_1[0]
        [0.0, 4.0]

        >>> test_1[1]
        [0.0, 1.0, 2.0, 3.0, 4.0]

        >>> ts_4_4.fill_2s_3s()
        >>> ts_4_4.pulses
        [4.0, 2.0, 1.0]

        >>> test_2 = ts_4_4.to_start_hierarchy()
        >>> test_2[0]
        [0.0, 4.0]

        >>> test_2[1]
        [0.0, 2.0, 4.0]

        >>> test_2[2]
        [0.0, 1.0, 2.0, 3.0, 4.0]

        >>> ts_2_2 = TimeSignature(as_string="2/2")
        >>> ts_2_2.pulses
        [4.0, 2.0]

        >>> test_3 = ts_2_2.to_start_hierarchy()
        >>> test_3[0]
        [0.0, 4.0]

        >>> test_3[1]
        [0.0, 2.0, 4.0]

        >>> ts_2_2.fill_2s_3s() # no effect, unchanged
        >>> ts_2_2.pulses
        [4.0, 2.0]

        >>> test_4 = ts_2_2.to_start_hierarchy()
        >>> test_4[0]
        [0.0, 4.0]

        >>> test_4[1]
        [0.0, 2.0, 4.0]

        >>> ts_6_8 = TimeSignature(as_string="6/8")
        >>> ts_6_8.pulses
        [3.0, 0.5]

        >>> test_5 = ts_6_8.to_start_hierarchy()
        >>> test_5[0]
        [0.0, 3.0]

        >>> test_5[1]
        [0.0, 0.5, 1.0, 1.5, 2.0, 2.5, 3.0]

        >>> ts_6_8.fill_2s_3s()
        >>> ts_6_8.pulses
        [3.0, 1.5, 0.5]

        >>> test_6 = ts_6_8.to_start_hierarchy()
        >>> test_6[0]
        [0.0, 3.0]

        >>> test_6[1]
        [0.0, 1.5, 3.0]

        >>> test_6[2]
        [0.0, 0.5, 1.0, 1.5, 2.0, 2.5, 3.0]

        >>> ts_5_4 = TimeSignature(as_string="5/4")
        >>> ts_5_4.pulses
        [5.0, 1.0]

        >>> test_7 = ts_5_4.to_start_hierarchy()
        >>> test_7[0]
        [0.0, 5.0]

        >>> test_7[1]
        [0.0, 1.0, 2.0, 3.0, 4.0, 5.0]

        >>> ts_2_3_4 = TimeSignature(as_string="2+3/4")
        >>> ts_2_3_4.pulses # as before
        [5.0, 1.0]

        >>> test_8 = ts_2_3_4.to_start_hierarchy()
        >>> test_8[0]
        [0.0, 5.0]

        >>> test_8[1]
        [0.0, 2.0, 5.0]

        >>> test_8[2]
        [0.0, 1.0, 2.0, 3.0, 4.0, 5.0]

        """
        # 1. Basic elements: all periodic cycles from the full cycle to the `beat_type` level.
        pulses = (
            PulseLengths(  # TODO consistency wrt what is added to the class.
                pulse_lengths=self.pulses, cycle_length=self.cycle_length
            )
        )
        start_hierarchy = pulses.to_start_hierarchy()

        # 2. irregular beat layer, if applicable.
        if len(self.beats) > 1:  # not a regular pulse e.g., (2, 3)
            bp = BeatPattern(
                self.beats, self.beat_type
            )  # TODO consistency wrt what is added to the class.
            beat_starts = bp.beat_pattern_to_start_hierarchy()
            start_hierarchy.append(beat_starts)

            start_hierarchy = [
                list(i) for i in set(map(tuple, start_hierarchy))
            ]
            start_hierarchy.sort(key=len)
            # TODO consider move to StartTimeHierarchy?

        return start_hierarchy


# ------------------------------------------------------------------------------


class PulseLengths:
    def __init__(
        self,
        pulse_lengths: list[float],
        cycle_length: Optional[float] = None,
        include_cycle_length: bool = True,
    ):
        """
        Representation of fully periodic meter centred on the constituent pulse lengths.

        Parameters
        ----------
        pulse_lengths
            Any valid list of pulse lengths, e.g., [4, 2, 1].
        cycle_length
            Optional. If not provided, the cycle length is taken to be given by the longest pulse length.
        include_cycle_length
            Defaults to True. If True, when converting to starts, include the full cycle length in the list.

        """

        self.pulse_lengths = pulse_lengths
        self.pulse_lengths.sort(reverse=True)  # largest number first

        self.cycle_length = cycle_length
        if self.cycle_length is not None:
            if pulse_lengths[0] > self.cycle_length:
                raise ValueError(
                    f"The `pulse_length` {pulse_lengths[0]} is longer than the `cycle_length` ({self.cycle_length})."
                )
        else:
            self.cycle_length = float(pulse_lengths[0])

        self.start_hierarchy = None
        self.include_cycle_length = include_cycle_length

    def to_start_hierarchy(
        self,
        require_2_or_3_between_levels: bool = False,
    ):
        """
        Convert a list of pulse lengths into a corresponding list of lists.

        Gives start positions per metrical level.
        All values (pulse lengths, start positions, and cycle_length)
        are all expressed in terms of quarter length.

        That is, the user provides pulse lengths for each level of a
        metrical hierarchy, and the algorithm expands this into a hierarchy
        assuming equal spacing (aka “isochrony”).

        This does not work for (“nonisochronous”) pulse streams of varying duration
        in time signatures like 5/x, 7/x (e.g., the level of 5/4 with
        dotted/undotted 1/2 notes).

        It is still perfectly fine to use this for the pulse streams
        within those meters that are regular, equally spaced (“isochronous”)
        (e.g., the 1/4 note level of 5/4).

        The list of pulse lengths is handled internally in decreasing order,
        whatever the ordering in the argument.

        If `require_2_or_3_between_levels` is True (default), this functions
        checks that each level is either a 2 or 3 multiple of the next.

        By default, the cycle_length is taken by the longest pulse length.
        Alternatively, this can be user-defined to anything as long as it is

        1. longer than the longest pulse and
        2. if `require_2_or_3_between_levels` is True then exactly 2x or 3x
        longer.


        Parameters
        ----------
        require_2_or_3_between_levels
            Defaults to False.
            If True, raise a ValueError in the case of this condition not being met.

        Returns
        -------
        list
            Returns a list of lists with start positions by level.

        Examples
        --------

        >>> qsl = PulseLengths(pulse_lengths=[4, 2, 1, 0.5])
        >>> qsl.pulse_lengths
        [4, 2, 1, 0.5]

        >>> start_hierarchy = qsl.to_start_hierarchy()
        >>> start_hierarchy[0]
        [0.0, 4.0]

        >>> start_hierarchy[1]
        [0.0, 2.0, 4.0]

        >>> start_hierarchy[2]
        [0.0, 1.0, 2.0, 3.0, 4.0]

        >>> start_hierarchy[3]
        [0.0, 0.5, 1.0, 1.5, 2.0, 2.5, 3.0, 3.5, 4.0]

        """

        if require_2_or_3_between_levels:  # TODO consider refactor
            for level in range(len(self.pulse_lengths) - 1):
                if self.pulse_lengths[level] / self.pulse_lengths[
                    level + 1
                ] not in [
                    2,
                    3,
                ]:
                    raise ValueError(
                        "The proportion between consecutive levels is not 2 or 3 in "
                        f"this case: {self.pulse_lengths[level]}:{self.pulse_lengths[level + 1]}."
                    )

        start_list = []

        for pulse_length in self.pulse_lengths:
            starts = self.one_pulse_to_start_hierarchy_list(pulse_length)
            start_list.append(starts)

        self.start_hierarchy = start_list
        return start_list

    def one_pulse_to_start_hierarchy_list(
        self,
        pulse_length: float,
    ):
        """
        Convert a single pulse length and cycle length into a list of starts.
        All expressed in quarter length.

        Note:
        A maximum of 4 decimal places is hardcoded.
        This is to avoid floating point errors or the need for one line of numpy (np.arange)
        in a module that doesn't otherwise use it.
        4 decimal points should be sufficient for all realistic use cases.

        Parameters
        --------
        pulse_length
            The quarter length of the pulse (note: must be shorter than the
            `cycle_length`).

        Examples
        --------

        >>> pls = PulseLengths(pulse_lengths=[4, 2, 1, 0.5], cycle_length=4)
        >>> pls.pulse_lengths
        [4, 2, 1, 0.5]

        >>> pls.one_pulse_to_start_hierarchy_list(1)
        [0.0, 1.0, 2.0, 3.0, 4.0]

        >>> pls = PulseLengths(pulse_lengths=[4, 2, 1, 0.5], cycle_length=4, include_cycle_length=False)
        >>> pls.one_pulse_to_start_hierarchy_list(1)
        [0.0, 1.0, 2.0, 3.0]

        """
        starts = []
        count = 0
        while count < self.cycle_length:
            starts.append(round(float(count), 4))
            count += pulse_length

        if self.include_cycle_length:
            starts.append(round(float(count), 4))

        return starts

    def to_array(self):
        """
        Create an array with the levels as rows,
        and the entries as pulse length values,
        where those pulse lengths start, and 0 otherwise.

        Examples
        --------

        >>> pls = PulseLengths(pulse_lengths=[4, 2, 1, 0.5], cycle_length=4)
        >>> pls.to_array()
        array([[4. , 0. , 0. , 0. , 0. , 0. , 0. , 0. ],
               [2. , 0. , 0. , 0. , 2. , 0. , 0. , 0. ],
               [1. , 0. , 1. , 0. , 1. , 0. , 1. , 0. ],
               [0.5, 0.5, 0.5, 0.5, 0.5, 0.5, 0.5, 0.5]])

        """
        tatum = self.pulse_lengths[-1]
        linear_positions = int(self.cycle_length / tatum)
        granular_row = np.full(
            shape=(1, linear_positions), fill_value=tatum, dtype=np.float64
        )
        symbolic_pulse_length_array = np.array(granular_row)
        for pulse in self.pulse_lengths[-2::-1]:  # TODO check len?
            multiple_of_tatum = int(pulse / tatum)
            divider_of_cycle = int(linear_positions / multiple_of_tatum)
            proto_row = np.zeros((1, multiple_of_tatum))
            proto_row[0, 0] = pulse
            row = proto_row
            for i in range(divider_of_cycle - 1):
                row = np.append(row, proto_row)
            symbolic_pulse_length_array = np.vstack(
                (row, symbolic_pulse_length_array)
            )

        return symbolic_pulse_length_array


class BeatPattern:
    """
    Encoding only the part of a metrical structure identified as the beat pattern.

    Parameters
    --------
    beat_list
        An ordered list of the beat types.
    beat_type
        The lower value of a time signature to set the pulse value.
    """

    def __init__(
        self,
        beat_list: tuple[int, ...],
        beat_type: int,
    ):

        self.beat_list = beat_list
        self.beat_type = beat_type
        self.start_time_hierarchy = self.beat_pattern_to_start_hierarchy()

    def beat_pattern_to_start_hierarchy(
        self, include_cycle_length: bool = True
    ) -> list:
        """
        Converts a list of beats
        like [2, 2, 2]
        or [3, 3]
        or indeed
        [6, 9]
        into a list of within-cycle starting positions, as defined relative
        to the start of the cycle.
        Basically, the list of beats functions like the time signature's
        so-called “numerator”,
        so for instance, `[2, 2, 3]` with the denominator `4` is a kind of 7/4.
        This equates to starting positions of
        `[0.0, 2.0, 4.0, 7.0]`.

        Parameters
        --------
        include_cycle_length
            If True (default) then each level ends with the full cycle length
            (i.e., the start of the next cycle).

        Examples
        --------

        >>> bp = BeatPattern((2, 2, 3), 4)
        >>> bp.beat_pattern_to_start_hierarchy()
        [0.0, 2.0, 4.0, 7.0]

        >>> bp.beat_pattern_to_start_hierarchy(include_cycle_length = False)
        [0.0, 2.0, 4.0]

        """
        starts = [0.0]  # always float, always starts at zero
        count = 0
        for beat_val in self.beat_list:
            count += beat_val
            this_start = count * 4 / self.beat_type
            starts.append(this_start)

        if include_cycle_length:  # include last value
            return starts
        else:
            return starts[:-1]


# -----------------------------------------------------------------------------

if __name__ == "__main__":
    import doctest

    doctest.testmod()



"""Test suite for functions in amads.time.notedensity"""

import pytest

from amads.core.basics import Note, Part, Score
from amads.core.timemap import TimeMap
from amads.time.notedensity import note_density


def test_notedensity_empty_notes():
    """Empty score should return 0.0"""
    score = Score.from_melody([])
    result = note_density(score, timetype="quarters")
    assert result == 0.0


def test_notedensity_seconds_score_seconds_timetype():
    """timetype is seconds, score.units_are_seconds = True"""
    # Create a score in seconds
    score = Score()
    score.convert_to_seconds()
    part = Part(parent=score)
    Note(parent=part, onset=0.0, duration=0.5, pitch=60)
    Note(parent=part, onset=0.5, duration=0.5, pitch=62)
    Note(parent=part, onset=1.0, duration=0.5, pitch=64)

    result = note_density(score, timetype="seconds")
    # (3-1) / 1.0 = 2.0
    assert result == 2.0


def test_notedensity_quarters_score_seconds_timetype():
    """timetype is quarters, score.units_are_seconds = True"""
    # Create a score in seconds with time_map
    score = Score()
    score.convert_to_seconds()
    score.time_map = TimeMap(qpm=120)
    part = Part(parent=score)
    Note(parent=part, onset=0.0, duration=0.5, pitch=60)
    Note(parent=part, onset=0.5, duration=0.5, pitch=62)
    Note(parent=part, onset=1.0, duration=0.5, pitch=64)

    result = note_density(score, timetype="quarters")
    # 0.0s = 0.0q, 1.0s = 2.0q (at 120 BPM)
    # duration = 2.0 - 0.0 = 2.0 quarters
    # density = (3-1) / 2.0 = 1.0
    assert result == 1.0


def test_notedensity_invalid_timetype():
    """Invalid timetype"""
    score = Score.from_melody([60, 62, 64])
    with pytest.raises(ValueError) as exc_info:
        note_density(score, timetype="invalid")
    assert "Invalid timetype: invalid. Use 'quarters' or 'seconds'." in str(
        exc_info.value
    )


def test_notedensity_duration_zero():
    """duration is zero"""
    score = Score()
    part = Part(parent=score)
    Note(parent=part, onset=0.0, duration=1.0, pitch=60)
    Note(parent=part, onset=0.0, duration=1.0, pitch=62)
    Note(parent=part, onset=0.0, duration=1.0, pitch=64)

    result = note_density(score, timetype="quarters")
    assert result == 0.0


def test_notedensity_default_timetype():
    """make sure that the default timetype is quarters"""
    score = Score.from_melody([60, 62, 64, 65])
    result = note_density(score)  # default timetype='quarters'
    assert result == 1.0



"""
Partimenti from the literature (broadly Gjerdingen)

Organized in simple dicts with keys for:

- `"name"`: e.g., `"Romanesca"`,
- `"when"`: e.g., `"Opening"`,
- `"strong_first"`: bool. True if the pattern starts strong-weak, False for the opposite. (They all alternate)
- `"melody"`: the melodic part expressed in scale degrees (int 1--7).
- `"bass"`: the bass part expressed in scale degrees (int 1--7).
- `"figures"`: harmony as expressed in figured bass (see toolkit issue #8).
- `"note"`: optional.

Ongoing considerations for the encoding:
- figure numbers: encoding figures is a very pervasive issue. This "X,Y" format is interoperable with some libraries.
- rhythmic differences between components. This "X-Y" preserves the N stages and metrical sides.

<small>**Author**: Mark Gotham, 2019</small>
"""

__author__ = "Mark Gotham"


# ------------------------------------------------------------------------------

# Opening

romanesca = {
    "name": "Romanesca",
    "when": "Opening",
    "strong_first": True,
    "melody": [1, 5, 1, 1],
    "bass": [1, 7, 6, 3],
    "figures": [5, 6, 5, 6],
}
do_re_mi = {
    "name": "Do-Re-Mi",
    "when": "Opening",
    "strong_first": True,
    "melody": [1, 2, 3],
    "bass": [1, 7, 1],
    "figures": [5, 6, 5],
    "note": "invertible",
}
sol_fa_mi = {
    "name": "Sol-Fa-Mi",
    "when": "Opening",
    "strong_first": False,
    "melody": [5, 4, 4, 3],
    "bass": [1, 2, 7, 1],
    "figures": [5, 5, "6,5", 5],
}
meyer = {
    "name": "Meyer",
    "when": "Opening",
    "strong_first": False,
    "melody": [1, 7, 4, 3],
    "bass": [1, 2, 7, 1],
    "figures": [5, "6,4,3", "6,5", 5],
}
aprile = {
    "name": "Aprile",
    "when": "Opening",
    "strong_first": True,
    "melody": [1, 7, 2, 1],
    "bass": [1, 2, 7, 1],
    "figures": [5, "6,4,3", "6,5", 5],
    "note": "Happy birthday!",
}
jupiter = {
    "name": "Jupiter",
    "when": "Opening",
    "strong_first": True,
    "melody": [1, 2, 4, 3],
    "bass": [1, 7, 5, 1],
    "figures": [5, 6, 5, 5],
}
pastorella = {
    "name": "Pastorella",
    "when": "Opening",
    "strong_first": True,
    "melody": [3, 2, 4, 3],
    "bass": [1, 5, 5, 1],
    "figures": [5, 6, 5, 5],
    "note": "Thirds",
}


# Answer/Process/Transition

prinner = {
    "name": "Prinner",
    "when": "Answer/Process/Transition",
    "strong_first": True,
    "melody": [6, 5, 4, 3],
    "bass": [4, 3, 2, 1],
    "figures": [5, 6, "7-6", 5],
}
modulating_prinner = {
    "name": "Modulating Prinner",
    "when": "Answer/Process/Transition, e.g. end of A",
    "strong_first": True,
    "melody": [3, 2, 1, 7],
    "bass": [8, 7, 6, 5],
    "figures": [5, 6, "7-#6", 5],
}
fonte = {
    "name": "Fonte",
    "when": "Answer/Process/Transition, e.g. start of B",
    "strong_first": False,
    "melody": [5, 4, 4, 3],
    "bass": ["#1", 2, 7, 1],
    "figures": ["6,5", 5, "6,5", 5],
    "note": "Cycle of 5ths. 6,5 due to melody",
}
monte = {
    "name": "Monte",
    "when": "Answer/Process/Transition, e.g. start of B",
    "strong_first": False,
    "melody": ["1-b7", "6", "2-1", "7"],
    "bass": [3, 4, "#4", 5],
    "figures": [6, 5, 6, 5],
}
ponte = {
    "name": "Ponte",
    "when": "Answer/Process/Transition",
    "strong_first": True,
    "melody": [5, 7, 2],
    "bass": [5, 5, 5],
    "figures": [5, 7, 7],
}


# "Pre-Cadential" (process / transition continued):

fenaroli = {
    "name": "Fenaroli",
    "when": "Pre-Cadential",
    "strong_first": True,
    "melody": [4, 3, 7, 1],
    "bass": [7, 1, 2, 3],
    "figures": [6, 5, 6, 6],
}
indugio = {
    "name": "Indugio",
    "when": "Pre-Cadential",
    "strong_first": True,
    "melody": [2, 4, 6, 1, 7],
    "bass": [4, 4, 4, "4#", 5],
    "figures": ["6,5", "6,5", "6,5", "6,5", 5],
}
passo_indietro = {
    "name": "Passo Indietro",
    "when": "Pre-Cadential",
    "strong_first": True,
    "melody": [7, 1],
    "bass": [4, 3],
    "figures": ["6,4,2", 6],
    "note": "Step back, before a significant cadence",
}
deceptive = {
    "name": "Deceptive Cadence",
    "when": "Pre-Cadential",
    "strong_first": False,
    "melody": [1, 2, 2, 1],
    "bass": [3, 4, 5, 6],
    "figures": [6, "6,5", 5, 5],
}
evaded = {
    "name": "Evaded Cadence",
    "when": "Pre-Cadential",
    "strong_first": False,
    "melody": [1, 2, 2, 1],
    "bass": [3, 4, 5, 3],
    "figures": [6, "6,5", 5, 6],
}


# Cadential

cadenza_semplice = {
    "name": "Cadenza Semplice",
    "when": "Cadence",
    "strong_first": False,
    "melody": [1, 2, 2, 1],
    "bass": [3, 4, 5, 1],
    "figures": [6, "6,5", 5, 5],
}
cadenza_composta = {
    "name": "Cadenza Composta",
    "when": "Cadence",
    "strong_first": True,
    "melody": [1, 2, 3, 2, 1],
    "bass": [3, 4, 5, 5, 1],
    "figures": [6, "6,5", "6,4", 7, 5],
}
cadenza_doppia = {
    "name": "Cadenza Doppia",
    "when": "Cadence",
    "strong_first": True,
    "melody": [4, 3, 2, 2, 1],
    "bass": [5, 5, 5, 5, 1],
    "figures": [5, "6,4", 4, 3, 5],
    "relatedTo": ["Comma", "Complete"],
}
complete = {
    "name": "Complete Cadence",
    "when": "Cadence",
    "strong_first": False,
    "melody": [3, 2, 2, 1],
    "bass": [1, 4, 5, 1],
    "figures": [5, "6,5", 5, 5],
}


# More Cadences (partial / incomplete)

comma = {
    "name": "Comma",
    "when": "Cadence",
    "strong_first": False,
    "melody": [4, 3],
    "bass": [7, 1],
    "figures": ["6,5", 5],
}
converging = {
    "name": "Converging Cadence",
    "when": "Cadence",
    "strong_first": False,
    "melody": [3, 2, 1, 7],
    "bass": ["3", "4", "#4", "5"],
    "figures": [6, "6,5", "6,5", 5],
}


# Post Cadence

quiescenza = {
    "name": "Quiescenza",
    "when": "Post-Cadential",
    "strong_first": False,
    "melody": ["b7", "6", "7", "1"],
    "bass": [1, 1, 1, 1],
    "figures": ["b7", "6,4", "7,4,2", 5],
}


all = [
    aprile,
    cadenza_doppia,
    cadenza_semplice,
    comma,
    converging,
    deceptive,
    do_re_mi,
    evaded,
    fonte,
    fenaroli,
    indugio,
    jupiter,
    meyer,
    modulating_prinner,
    monte,
    passo_indietro,
    pastorella,
    ponte,
    prinner,
    quiescenza,
    romanesca,
    sol_fa_mi,
]



# pitchmean.py - compute mean or duration-weighted mean of pitch
#

from amads.core.basics import Note


def pitch_mean(score, weighted=False):
    """Compute the mean pitch or mean pitch weighted by duration (in quarters)

    Parameters
    ----------
    score : Score
        The pitch mean is computed for all pitches in the score. Groups of
        two or more tied notes are counted as one pitch occurrence.
    weighted : bool
        If true, pitches are weighted by their durations.
    """
    sum = 0
    count = 0
    if weighted:  # no need to merge tied notes. Rather than merging tied
        # notes, we use pitch * tied_duration = = pitch * (dur_1 + dur_2)
        # = pitch * dur_1 + pitch * dur_2, so we can just treat the
        # components of tied notes separately.
        for note in score.find_all(Note):
            sum += note.key_num * note.duration
            count += note.duration
    else:
        # Our problem is that we want to count only the first of any tied-note
        # group. We could use merge_tied_notes() to do this, but it is work to
        # make a new scoore and copy all notes. Instead, we keep a set of
        # tied-to notes and ignore any note encountered that is in the set.
        tied_to = set()
        for note in score.find_all(Note):
            if note.tie is not None:
                tied_to.add(note.tie)
            if note not in tied_to:
                sum += note.key_num
                count += 1
    return (sum / count) if sum > 0 else 0



"""
Basic functionality for transforming pitch lists
expressed as integers (MIDI numbers or pitch classes)
through transposition, inversion, retrograde, rotation, and more.

Most apply equally to any pitch class sequence
and can therefore be use in melody and harmony settings.
"""

from typing import Iterable

# ------------------------------------------------------------------------------


def transpose_by(
    pitches: Iterable, semitones: int, mod_12: bool = True
) -> list:
    """
    Transposes a list of pitches by an interval of size
    set by the value of `semitones`.

    Parameters
    ----------
    pitches
        Any list or tuple of integers representing pitches as MIDI numbers or pitch classes.
    semitones
        How far to transpose, expressed in semitones (1 per MIDI note).
    mod_12
        If True, return values modulo 12 (necessary for pitch class sets, not for MIDI numbers)

    Returns
    -------
    list
        A new list of the same length as the input.

    Examples
    --------

    >>> transpose_by([0, 1, 2, 3,], 16, mod_12=True)
    [4, 5, 6, 7]

    >>> transpose_by([0, 1, 2, 3,], 16, mod_12=False)
    [16, 17, 18, 19]

    """
    result = []
    for pitch in pitches:
        transposed = pitch + semitones
        if mod_12:
            transposed %= 12
        result.append(transposed)
    return result


def transpose_to(
    pitches: Iterable, start: int = 0, mod_12: bool = True
) -> list:
    """
    Transpose a list of pitch classes to start on 0 (by default), or
    any another number set by the value of `start`.

    Parameters
    ----------
    pitches
        Any list or tuple of integers representing pitches as MIDI numbers or pitch classes.
    start
        The first number of the new list.
    mod_12
        If True, return values modulo 12 (necessary for pitch class sets, not for MIDI numbers)

    Returns
    -------
    list
        A new list of the same length as the input.

    Examples
    --------

    >>> transpose_to([0, 1, 2, 3,], 16, mod_12=True)
    [4, 5, 6, 7]

    >>> transpose_to([0, 1, 2, 3,], 16, mod_12=False)
    [16, 17, 18, 19]

    """
    difference = start - pitches[0]
    result = []
    for pitch in pitches:
        transposed = pitch + difference
        if mod_12:
            transposed %= 12
        result.append(transposed)
    return result


def retrograde(pitches: Iterable) -> list:
    """
    Retrograde (reverse) a list of pitches.

    Parameters
    ----------
    pitches
        Any list or tuple of integers representing pitches as MIDI numbers or pitch classes.

    Returns
    -------
    list
        A new list of the same length as the input.

    Examples
    --------

    >>> retrograde([2, 6, 9])
    [9, 6, 2]

    """
    result = pitches[::-1]  # to create a copy
    return result


def invert(
    pitches: Iterable, use_first_not_0: bool = True, mod_12: bool = True
) -> list:
    """
    Invert a list of pitch classes around a specified pitch: the starting pitch or 0.

    Parameters
    ----------
    pitches
        Any list or tuple of integers representing pitches as MIDI numbers or pitch classes.
    use_first_not_0
        If true, use the first number of the list as the centre of the inversion.
    mod_12
        If True, return values modulo 12 (necessary for pitch class sets, not for MIDI numbers)

    Returns
    -------
    list
        A new list of the same length as the input.

    Examples
    --------
    >>> invert([5, 6, 7])
    [5, 4, 3]

    >>> invert([5, 6, 7], use_first_not_0=False)
    [7, 6, 5]

    """
    origin = pitches[0] if use_first_not_0 else 0
    result = []
    for pitch in pitches:
        inverted = 2 * origin - pitch
        if mod_12:
            inverted %= 12
        result.append(inverted)

    return result


def pitches_to_intervals(
    pitches: Iterable, wrap: bool = False, mod_12: bool = True
) -> list:
    """
    Get the interval succession of a list of pitches.

    Parameters
    ----------
    pitches
        Any list or tuple of integers representing pitches as MIDI numbers or pitch classes.
    wrap
        If true, include the interval from the last element to the first in addition.
    mod_12
        If True, return values modulo 12 (necessary for pitch class sets, not for MIDI numbers)

    Returns
    -------
    list
        A new list of the same length as the input (if wrap), otherwise, one less.

    Examples
    --------
    >>> pitches_to_intervals([0, 2, 5])
    [2, 3]

    >>> pitches_to_intervals([0, 2, 5], wrap=True)
    [2, 3, 7]

    """
    intervals = []
    if wrap:
        pitches += [pitches[0]]
    for i in range(1, len(pitches)):
        i = pitches[i] - pitches[i - 1]
        if mod_12:
            i %= 12
        intervals.append(i)

    return intervals


def rotate(pitches: Iterable, steps: int = 1) -> list:
    """
    Rotates a list of pitch classes through N steps (i.e. starts on the Nth element).

    Should be called on an integer less than the length of the pitch list.
    If called on a larger integer, the value modulo the length of the pitch list.
    (e.g. 15 becomes 3 for a pitch list of length 12).

    Parameters
    ----------
    pitches
        Any list or tuple of integers representing pitches as MIDI numbers or pitch classes.
    steps
        If true, include the interval from the last element to the first in addition.

    Returns
    -------
    list
        A new list of the same length as the input.

    Examples
    --------
    >>> rotate([0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11], 4)
    [4, 5, 6, 7, 8, 9, 10, 11, 0, 1, 2, 3]


    """
    list_len = len(pitches)
    if steps > list_len:
        steps = steps % list_len
    result = pitches[steps:] + pitches[:steps]
    return result


def every_nth(pitches: list, start_index: int = 0, step_size: int = 5) -> list:
    """
    Cycle through a list of pitches
    with a step size of n mod the length of the list

    By default,
    start at index 0 and iterate 12 times (0-12),
    though both the start index and the range() are settable arguments,
    hence this equivlance:

    Parameters
    ----------
    pitches
        Any list or tuple of integers representing pitches as MIDI numbers or pitch classes.
    start_index
        The index position in the list to start at.
    step_size
        The gap between successive elements.

    Returns
    -------
    list
        A new list of the same length as the input.

    Examples
    --------

    >>> test_pitches = [11, 9, 6, 8, 5, 10, 0, 4, 3, 1, 2, 7]
    >>> every_nth(test_pitches, step_size=5)
    [11, 10, 2, 8, 3, 9, 0, 7, 5, 1, 6, 4]

    >>> every_nth(test_pitches, step_size=5, start_index=5)
    [10, 2, 8, 3, 9, 0, 7, 5, 1, 6, 4, 11]

    """
    list_len = len(pitches)
    out_pitches = []
    for i in range(list_len):
        out_pitches.append(pitches[(start_index + i * step_size) % list_len])
    return out_pitches


# ------------------------------------------------------------------------------

if __name__ == "__main__":
    import doctest

    doctest.testmod()



"""Functions for music data input."""

__author__ = "Roger B. Dannenberg"

import pathlib
import tempfile
import urllib.request
import warnings
from typing import Callable, Optional

from amads.core.basics import Score
from amads.io.writescore import _suffix_to_format

# This module, readscore, is regarded as a singleton class with
# the following attributes:

preferred_midi_reader: str = "pretty_midi"  # subsystem for MIDI file input
preferred_xml_reader: str = "music21"  # subsystem for MusicXML input
preferred_kern_reader: str = "music21"  # subsystem for Kern input
preferred_mei_reader: str = "music21"  # subsystem for mei input
reader_warning_level: str = "default"  # controls verbocity of warnings
#     from input processing
_last_used_reader: Optional[Callable] = None  # See last_used_reader()

valid_score_extensions = list(_suffix_to_format.keys())
valid_score_extensions.remove(".pdf")
valid_score_extensions.remove(".ly")  # write-only extensions

allowed_subsystems = {
    "midi": ["music21", "pretty_midi"],
    "musicxml": ["music21", "partitura"],
    "kern": ["music21", "partitura"],
    "mei": ["music21", "partitura"],
}

# mapping from preference strings to subsystem callables
_subsystem_map = {
    "music21": ("amads.io.m21_import", "music21_import"),
    "pretty_midi": ("amads.io.pm_midi_import", "pretty_midi_import"),
    "partitura": ("amads.io.pt_import", "partitura_import"),
}


def set_preferred_midi_reader(reader: str) -> str:
    """
    Set a (new) preferred MIDI reader.

    Returns the previous reader preference. The current preference is stored
    in `amads.io.reader.preferred_midi_reader`.

    Parameters
    ----------
    reader : str
        The name of the preferred MIDI reader; "music21" or "pretty_midi".

    Returns
    -------
    str
        The previous name of the preferred MIDI reader.

    Raises
    ------
    ValueError
        If an invalid reader is provided.
    """
    global preferred_midi_reader
    allowed = ["music21", "pretty_midi"]
    if reader not in allowed_subsystems["midi"]:
        raise ValueError(f"Invalid MIDI reader. Must be one of {allowed}")

    previous = preferred_midi_reader
    preferred_midi_reader = reader
    return previous


def set_preferred_xml_reader(reader: str) -> str:
    """
    Set a (new) preferred XML reader.

    Returns the previous reader preference. The current preference is stored
    in `amads.io.reader.preferred_xml_reader`.

    Parameters
    ----------
    reader : str
        The name of the preferred XML reader. Can be "music21" or "partitura".

    Returns
    -------
    str
        The previous name of the preferred XML reader.

    Raises
    ------
    ValueError
        If an invalid reader is provided.
    """
    global preferred_xml_reader
    allowed = allowed_subsystems["musicxml"]
    if reader not in allowed:
        raise ValueError(f"Invalid XML reader. Must be one of {allowed}")

    previous = preferred_xml_reader
    preferred_xml_reader = reader
    return previous


def set_preferred_kern_reader(reader: str) -> str:
    """
    Set a (new) preferred Kern reader.

    Returns the previous reader preference. The current preference is stored
    in `amads.io.reader.preferred_kern_reader`.

    Parameters
    ----------
    reader : str
        The name of the preferred Kern reader. Can be "music21" or "partitura".

    Returns
    -------
    str
        The previous name of the preferred Kern reader.

    Raises
    ------
    ValueError
        If an invalid reader is provided.
    """
    global preferred_kern_reader
    allowed = allowed_subsystems["kern"]
    if reader not in allowed:
        raise ValueError(f"Invalid Kern reader. Must be one of {allowed}")

    previous = preferred_kern_reader
    preferred_kern_reader = reader
    return previous


def set_preferred_mei_reader(reader: str) -> str:
    """
    Set a (new) preferred MEI reader.

    Returns the previous reader preference. The current preference is stored
    in `amads.io.reader.preferred_mei_reader`.

    Parameters
    ----------
    reader : str
        The name of the preferred MEI reader. Can be "music21" or "partitura".

    Returns
    -------
    str
        The previous name of the preferred MEI reader.

    Raises
    ------
    ValueError
        If an invalid reader is provided.
    """
    global preferred_mei_reader
    allowed = allowed_subsystems["mei"]
    if reader not in allowed:
        raise ValueError(f"Invalid MEI reader. Must be one of {allowed}")

    previous = preferred_mei_reader
    preferred_mei_reader = reader
    return previous


def set_reader_warning_level(level: str) -> str:
    """
    Set the warning level for `readscore` functions.

    The translation from music data files to AMADS is not always well-defined
    and may involve intermediate representations using Music21, Partitura or
    others. Usually, warnings are produced when there is possible data loss or
    ambiguity, but these can be more annoying than informative. The warning
    level can be controlled using this function, which applies to all file
    formats.

    The current warning level is stored in
    `amads.io.reader.reader_warning_level`.

    Parameters
    ----------
    level : str
        The warning level to set.
        Options are "none", "low", "default", "high".

        - "none" will suppress all warnings during `read_score()`
          and also suppresses notice of reader subsystem and file name.
        - "low" will print one notice if there are any warnings.
        - "default" will obey environment settings to control warnings.
        - "high" will print all warnings during `read_score()`, overriding
            environment settings.

    Returns
    -------
    str
        Previous warning level.

    Raises
    ------
    ValueError
        If an invalid warning level is provided.
    """
    global reader_warning_level
    allowed = ["none", "low", "default", "high"]
    if level not in allowed:
        raise ValueError(f"Invalid warning level. Must be one of {allowed}")

    previous = reader_warning_level
    reader_warning_level = level
    return previous


def _check_for_subsystem(
    format: str,
) -> tuple[
    Optional[Callable[[str, str, bool, bool, bool, bool], Score]], Optional[str]
]:
    """
    Check if the preferred reader is available.

    We support:
    `music21` for midi and xml,
    `partitura` for xml,
    and
    `PrettyMIDI` for midi.

    Partitura has basic MIDI import functionality, but is unsupported here
    because when it reads in a score it has no MIDI velocity
    and when it reads in a performance it has no tempo track, key signature, etc.

    Parameters
    ----------
    format : str
        The type of file to read: 'midi', 'musicxml', 'kern', or 'mei'.

    Returns
    -------
    tuple[Optional[Callable], Optional[str]]
        The import function if available, None otherwise.
    """
    preferred_reader = {
        "midi": preferred_midi_reader,
        "musicxml": preferred_xml_reader,
        "kern": preferred_kern_reader,
        "mei": preferred_mei_reader,
    }.get(format)

    if not preferred_reader:
        return None, preferred_reader

    try:
        if (
            preferred_reader not in _subsystem_map
            or preferred_reader not in allowed_subsystems[format]
        ):
            raise ValueError(
                f"Preferred reader '{preferred_reader}' not supported for "
                f"{format} import."
            )

        module_name, func_name = _subsystem_map[preferred_reader]
        module = __import__(module_name, fromlist=[func_name])
        return getattr(module, func_name), preferred_reader
    except Exception as e:
        print(f"Error importing {preferred_reader} for {format} files: {e}")
    return None, preferred_reader


def _import_score(
    filename: str,
    format: str,
    flatten: bool = False,
    collapse: bool = False,
    show: bool = False,
    group_by_instrument: bool = True,
) -> Score:
    """Import a score file

    <small>**Author**: Roger B. Dannenberg</small>
    """
    global _last_used_reader
    import_fn, preferred_reader = _check_for_subsystem(format)
    if import_fn is not None:
        _last_used_reader = import_fn
        if reader_warning_level != "none":
            print(
                f"Reading {filename} using {format} reader "
                f"file={import_fn.__name__}."
            )
        return import_fn(
            filename, format, flatten, collapse, show, group_by_instrument
        )
    else:
        raise Exception(
            "Could not find a MusicXML import function. "
            f"Preferred subsystem is {preferred_reader}"
        )


def read_score(
    filename: str,
    flatten: bool = False,
    collapse: bool = False,
    show: bool = False,
    format: Optional[str] = None,
    group_by_instrument: bool = True,
) -> Score:
    """Read a file with the given format, 'musicxml', 'midi', 'kern', 'mei'.

    If format is None (default), the format is based on the filename
    extension, which can be 'musicxml', 'mid', 'midi', 'smf', 'kern',
    or 'mei'. (Valid extensions are in
    `amads.io.readscore.valid_score_extensions`.)

    <small>**Author**: Roger B. Dannenberg</small>

    Parameters
    ----------
    filename : str
        The path (relative or absolute) to the music file.
        Can also be an URL.
    flatten : bool
        The returned score will be flat (Score, Parts, Notes).
    collapse: bool
        If collapse and flatten, the parts will be merged into one.
    show : bool
        Print a text representation of the data.
    format: string
        One from among limited standard options (e.g.,
        `'musicxml'`, `'midi'`, `'kern'`, `'mei'`)
    group_by_instrument : bool
        If True (default), when the underlying reader (e.g. for "pretty_midi",
        "music21" or "partitura") reads Parts with the same instrument, their
        content will be grouped into a single part. This means that if
        `flatten`, then parts with the same instrument will be merged into a
        single part. If `flatten` is False, then the staffs of parts with the
        same instrument will be grouped within a single part.
        If `group_by_instrument` is False, the parts read in by the underlying
        reader will be preserved as separate parts. `group_by_instrument` is
        True by default so that when reading Piano scores with separate treble
        and bass staffs, the resulting AMADS Score will generally have a single
        Piano part with two staffs. A score for Piano and Violin will generally
        have two parts, one for Piano and one for Violin, as opposed to three
        parts (Piano-Treble, Piano-Bass, Violin). On the other hand, a score for
        two Violins might be represented a one part with two staffs by default,
        but setting `group_by_instrument` to False will more likely keep the
        two Violin parts separate. Unfortunately, exact behavior depends on the
        underlying reader, MIDI track names, and/or MusicXML structure and
        naming.

    Returns
    -------
    Score
        The imported score

    Raises
    ------
    ValueError
        If the format is unknown or not implemented.

    Note on Incomplete First Measure
    --------------------------------
    In Music21, the first measure may be a partial measure containing
    an anacrusis (“pickup”). This is somewhat ambiguous and does not
    translate well to MIDI which is less expressive than MusicXML.

    Therefore, if the first measure read with Music21 is not a full
    measure, a rest is inserted and the remainder is shifted to
    form a full measure according to its time signature. Remaining
    measures are shifted in time accordingly and Score, Part and
    Staff durations are adjusted accordingly.

    General MIDI Import Notes
    -------------------------
    Each Standard MIDI File track corresponds to a Staff when
    creating a full AMADS Score. Everything is combined into one
    part when `flatten` and `collapse` are specified.

    AMADS assumes that instruments (midi program numbers) are fixed
    for each Staff (or Part in flat scores), and MIDI channels are
    not represented. The use of program change messages within a
    track to change the program are ignored, but may generate warnings.

    In general, AMADS instrument name corresponds to the MIDI track
    name, and MIDI program numbers are stored as `"midi_program"`
    in the `info` attribute of the Staff or Part corresponding to
    the track.

    MIDI files do not have a Part/Staff structure, but you can
    write multiple tracks with the same name. Both the `"music21"`
    and `"pretty_midi"` readers will group tracks with matching
    names as Staffs in a single Part. This may result in an
    unexpected Part/Staff hierarchy if tracks are not named or
    if tracks are named something like "Piano-Treble" and
    "Piano-Bass", which would produce two Parts as different
    instruments as opposed to one Part with two Staffs.

    Unless `flatten` or `collapse`, the MIDI file time signature
    information will be used to form Measures with Staffs, and
    Notes will be broken where they cross measure boundaries and
    then tied.  The default time signature is 4/4.

    Pretty MIDI Import Notes
    ------------------------
    If there is no program change in a file, the `"pretty_midi"`
    reader will use 0, and 0 will be stored as `"midi_program"`
    in the Part or Staff's `info` (see
    [get][amads.core.basics.Event.get] and
    [set][amads.core.basics.Event.set]).

    If there is no track name, the `Part.instrument` is derived
    from the track program number (defaults to zero).

    If the MIDI file track name is `"Unknown"`, the `Part.instrument`
    is set to None. This is because when the `"pretty_midi"` writer
    writes a part where `Part.instrument is None`, the name `"Unknown"`
    is used instead. Therefore, the reader will recreate the AMADS
    Part where `Part.instrument is None`.

    Pretty MIDI will not insert any KeySignature unless key signature
    meta-events are found.

    Music21 MIDI Import Notes
    -------------------------
    Music21 may infer a Clef and KeySignature even though MIDI
    does not even have a meta-event for clefs, and even if the
    MIDI file has no key signature meta-event.
    """
    if filename.startswith("http") or "://" in filename:
        with tempfile.NamedTemporaryFile(
            suffix=pathlib.Path(filename).suffix or ".tmp", delete=False
        ) as tmp_file:
            urllib.request.urlretrieve(filename, tmp_file.name)
            filename = tmp_file.name

    if format is None:
        ext = pathlib.Path(filename).suffix.lower()
        if ext not in [".pdf", ".ly"]:  # these are write-only extensions
            format = _suffix_to_format.get(ext)
        if not format:
            raise ValueError(
                f"Unsupported file extension: {ext}. "
                f"Valid extensions: {valid_score_extensions}"
            )

    # File format handling
    with warnings.catch_warnings(record=True) as w:
        warnings.simplefilter(
            "ignore" if reader_warning_level == "none" else "always"
        )

        score = _import_score(
            filename, format, flatten, collapse, show, group_by_instrument
        )

        # Warning handling
        if reader_warning_level == "low":
            if len(w) > 0:
                print(
                    f"Warning: {len(w)} warnings were generated in "
                    f"read_score({filename}).\n"
                    "  Use amads.io.readscore.set_reader_warning_level() "
                    "for more details."
                )
        else:  # "none", "default", or "high"
            for warning in w:
                print(
                    f"{warning.filename}:{warning.lineno}: "
                    f"{warning.category.__name__}: {warning.message}"
                )

        return score


def last_used_reader() -> Optional[str]:
    """Return the name of the last used reader function.

    Returns
    -------
    Optional[str]
        The name of the actual function used in the last call to `read_score`,
        or None if no reader has been used yet.
    """
    if _last_used_reader is not None:
        return _last_used_reader.__name__
    return None



"""
Basic properties of rhythms, which is to say 1D representations of musical events,
without measures, beats etc., and certainly no scores.

Broadly, this is for stand-alone functions clearly intended for
short, simple representations of rhythmic cycles,
and is not suitable for calling on scores, for instance.

This also includes some measures of rhythmic _complexity_
 (which is clearly not the same as syncopation, but related, and often studied together.

<small>**Author**: Mark Gotham</small>
"""

__author__ = "Mark Gotham"


from fractions import Fraction
from typing import Union

from amads.algorithms.gcd import integer_gcd_pair
from amads.core.vector_transforms_checks import (
    indicator_to_indices,
    indicator_to_interval,
)
from amads.core.vectors_sets import vector_to_multiset


def has_oddity_property(vector: Union[list[int], tuple[int, ...]]) -> bool:
    """
    Given a rhythm cycle (i.e., with the expectation of repetition) as a
    vector, check if it has Arom's "rhythmic-oddity" property:
    no two onsets partition the cycle into two equal parts.
    This is slightly confusing to get the right way around:
    the function returns `True` (i.e., yes, has the property)
    in the _absence_ of this equal division.

    Parameters
    ----------
    vector
        A vector for either the event positions in the cycle time span, or
        the beat pattern (sic, either).

    Returns
    -------
    bool
        True if the rhythm has the rhythmic-oddity property, False otherwise.

    Examples
    --------

    >>> son = (1, 0, 0, 1, 0, 0, 1, 0, 0, 0, 1, 0, 1, 0, 0, 0)
    >>> has_oddity_property(son)
    True

    >>> bembé = (1, 0, 1, 0, 1, 1, 0, 1, 0, 1, 0, 1)
    >>> has_oddity_property(bembé)
    False

    And here's a simple rhythm that does have the equal division:
    >>> has_oddity_property((1, 0, 1, 0, 1, 1, 1, 0, 1, 0, 1, 1))
    False

    Note that there does not need to be any further similarity between the two halves:
    >>> has_oddity_property((1, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 1))
    False

    """
    if not isinstance(vector, (list, tuple)):
        raise TypeError("The `vector` must be a list or tuple.")

    vector_length = len(vector)
    if vector_length % 2 != 0:
        return True  # By definition
    half_length = vector_length // 2

    indices = indicator_to_indices(vector)
    indices_set = set(indices)
    for i in indices:
        if (i + half_length) % vector_length in indices_set:
            return False

    return True


def _is_power_of_2(n: int) -> bool:
    """Return True if n is a positive power of 2."""
    return n > 0 and (n & (n - 1)) == 0


def _prev_power_of_2(x: int) -> int:
    """Return the largest power of 2 that is <= x."""
    return 1 << (x.bit_length() - 1)


def keith_via_toussaint(vector):
    """
    Although Keith's measure is described in terms of beats,
    it is inflexible to metric structure and fully defined by
    the onset pattern.

    Parameters
    ----------
    vector
        An indicator vector whose length must be a power of 2.

    Raises
    ------
    ValueError
        If the vector length is not a power of 2.

    Examples
    --------
    >>> son = [1, 0, 0, 1, 0, 0, 1, 0, 0, 0, 1, 0, 1, 0, 0, 0]
    >>> keith_via_toussaint(son)
    2

    """
    if not _is_power_of_2(len(vector)):
        raise ValueError(
            f"Vector length (currently {len(vector)}) must be a power of 2."
        )

    indices = indicator_to_indices(vector)  # Keith/Toussaint's `S`
    deltas = indicator_to_interval(
        vector, wrap=True
    )  # Keith/Toussaint also `delta`
    powers_of_2 = [
        _prev_power_of_2(x) for x in deltas
    ]  # Keith/Toussaint's big D
    count = 0
    for i in range(len(indices)):
        this_case = indices[i] / powers_of_2[i]
        if int(this_case) != this_case:
            count += 1
    return count


def has_deep_property(vector: Union[list[int], tuple[int, ...]]) -> bool:
    """
    So-called "Deep" rhythms have distinct numbers of each interval class
    among all (not-necessarily adjacent) intervals.
    See `indicator_to_interval` with the arguments `wrap=True`,
    `adjacent_not_all=False`.

    Examples
    --------

    >>> shiko = (1, 0, 0, 0, 1, 0, 1, 0, 0, 0, 1, 0, 1, 0, 0, 0)
    >>> indicator_to_interval(shiko, wrap=True, adjacent_not_all=False)
    (0, 2, 0, 3, 0, 4, 0, 1)

    Note the distinct numbers in the above.

    >>> has_deep_property(shiko)
    True

    The son clave is not deep, since multiple interval classes share the same count:

    >>> son = (1, 0, 0, 1, 0, 0, 1, 0, 0, 0, 1, 0, 1, 0, 0, 0)
    >>> has_deep_property(son)
    False

    """
    intervals = indicator_to_interval(vector, wrap=True, adjacent_not_all=False)
    non_zero_uses = [x for x in intervals if x != 0]
    return len(non_zero_uses) == len(set(non_zero_uses))


def off_beatness(vector: Union[list[int], tuple[int, ...]]) -> int:
    """
    The "off-beatness" measure records the number of events in a rhythmic cycle
    at positions which cannot fall on a regular beat division of the cycle.
    For a more formal definition, see `totatives`.

    This function expects an indicator vector (values of 0 or 1).
    Behaviour with non-indicator (weighted) vectors is undefined.

    Examples
    --------

    Gomez et al. explore 10 "canonical" 12-unit rhythms of which they
    find the Bembé notable for being "the most frequently used" and
    because it realizes the "highest value of off-beatness" among these 10.

    >>> bembé = (1, 0, 1, 0, 1, 1, 0, 1, 0, 1, 0, 1)
    >>> off_beatness(bembé)
    3

    Looking beyond these cases, the true highest value for a 12-unit
    cycle is 4 (using indices 1, 5, 7, 11), as shown in the minimal
    case here:

    >>> off_beatness((0, 1, 0, 0, 0, 1, 0, 1, 0, 0, 0, 1))
    4

    """
    t = set(totatives(len(vector)))
    positions = vector_to_multiset(vector)
    return sum(1 for item in t if item in positions)


def totatives(n):
    """
    Calculates the totatives of n, which are the positive integers less
    than n that are relatively prime to n.

    Parameters
    ----------
    n : int
        A positive integer. In the rhythmic case, this denotes cycle length.

    Returns
    -------
    list[int]
        A list of integers representing the totatives of n.
        This list is empty if n is less than or equal to 1.

    Examples
    --------
    >>> totatives(12)
    [1, 5, 7, 11]

    >>> len(totatives(12))
    4

    >>> totatives(16)
    [1, 3, 5, 7, 9, 11, 13, 15]

    >>> len(totatives(16))
    8

    """
    if n <= 1:
        return []

    return [i for i in range(1, n) if integer_gcd_pair(n, i) == 1]


def vector_to_onset_beat(
    vector: Union[list[int], tuple[int, ...]], beat_unit_length: int = 2
) -> tuple:
    """
    Map from a vector to onset beat data via `vector_to_multiset`.

    Parameters
    ----------
    vector
        An indicator vector.
        When representing a cycle with rotation,
        the final element should be a repeat of the first onset to mark the end of the cycle
        (hence vectors of length `cycle_length + 1` are common here).
    beat_unit_length
        The number of subdivisions per beat.
        Default is 2.

    Examples
    --------

    The son clave in 16 subdivisions.
    The 17th element marks the cycle end:

    >>> son = [1, 0, 0, 1, 0, 0, 1, 0, 0, 0, 1, 0, 1, 0, 0, 0, 1]
    >>> vector_to_onset_beat(vector=son, beat_unit_length=4)
    (Fraction(0, 1), Fraction(3, 4), Fraction(3, 2), Fraction(5, 2), Fraction(3, 1), Fraction(4, 1))

    """
    onsets = vector_to_multiset(vector)
    return tuple(Fraction(x, beat_unit_length) for x in onsets)


# ------------------------------------------------------------------------

if __name__ == "__main__":
    import doctest

    doctest.testmod()



import io
import math
from typing import List, Union, cast

import pytest

from amads.core.basics import (
    Chord,
    Clef,
    KeySignature,
    Measure,
    Note,
    Part,
    Pitch,
    Rest,
    Score,
    Staff,
)


def test_event_onset_none_raises():
    note = Note(parent=None, onset=None, duration=1.0)
    try:
        _ = note.onset
    except ValueError as e:
        assert str(e) == "Onset time is not set."
    else:
        assert False, "Expected ValueError when onset is None"


def test_from_melody_overlapping_notes():
    """Test that overlapping notes raise a ValueError."""
    with pytest.raises(ValueError) as exc_info:
        Score.from_melody(
            pitches=[60, 62],
            durations=2.0,  # half notes
            iois=1.0,  # but only 1 beat apart
        )
    assert (
        str(exc_info.value)
        == "Notes overlap: note 0 ends at 2.00 but note 1 starts at 1.00"
    )


def test_from_melody_iois_onsets():
    """Test that iois + onsets raise a ValueError."""
    with pytest.raises(ValueError) as exc_info:
        Score.from_melody(pitches=[60, 62], onsets=[0, 1], iois=1.0)
    assert str(exc_info.value) == "Cannot specify both iois and onsets"


def test_from_melody_onsets_len():
    """Test that onsets length problems raise a ValueError."""
    with pytest.raises(ValueError) as exc_info:
        Score.from_melody(pitches=[60, 62], onsets=[0, 1, 2])
    assert str(exc_info.value) == "onsets list must have same length as pitches"


def test_from_melody_list_lens():
    """Test that list length problems raise a ValueError."""
    with pytest.raises(ValueError) as exc_info:
        Score.from_melody(pitches=[60, 62], onsets=[0, 1], durations=[1, 1, 1])
    assert str(exc_info.value) == "All input lists must have the same length"


def test_from_melody_iois_len():
    """Test that iois list length problems raise a ValueError."""
    with pytest.raises(ValueError) as exc_info:
        Score.from_melody(pitches=[60, 62], iois=[0, 1])
    assert str(exc_info.value) == "iois list must have length len(pitches) - 1"


def test_from_melody_empty_pitches():
    """Test that an empty list of pitches creates a valid empty score."""
    score = Score.from_melody(pitches=[])
    assert score.duration == 0.0
    assert len(score.content) == 1  # should have one empty part
    # part should have no notes:
    assert len(cast(Part, score.content[0]).content) == 0


def test_invalid_note():
    melody: list[Union[Pitch, int, float, str]] = ["C", "H", "G"]
    with pytest.raises(ValueError):
        Score.from_melody(melody)


def test_copy_score():
    score = Score.from_melody(pitches=[60, 62, 64], durations=1.0)
    copied_score = cast(Score, score.copy())

    copied_note = cast(Note, cast(Part, copied_score.content[0]).content[0])
    assert isinstance(copied_note, Note)

    copied_part = copied_score.content[0]
    assert isinstance(copied_part, Part)

    assert copied_note.parent == copied_part
    assert copied_part.parent == copied_score

    assert copied_part.parent != score


def test_parent():
    score = Score.from_melody(pitches=[60, 62, 64], durations=1.0)

    part = score.content[0]
    assert isinstance(part, Part)
    assert part.part is None
    assert part.parent == part.score == score

    note = part.content[0]
    assert isinstance(note, Note)
    assert note.part == part
    assert note.score == score


def test_quantize():
    """See if when tied notes quantize to zero that the tied note
    is removed correctly.
    """
    # case 1: C tied to short C
    note1 = Note(duration=1, pitch="C4")
    note2 = Note(duration=0.125, pitch="C4")
    note1.tie = note2
    meas1 = Measure(note1, note2)
    staff = Staff(meas1)
    score = Score(Part(staff))

    # case 2: short D tied to long D
    note1 = Note(parent=meas1, onset=1.875, duration=0.125, pitch="D4")
    note2 = Note(parent=meas1, onset=2, duration=1, pitch="D4")
    note1.tie = note2

    # case 3: same as case 1 but F4 tied across measure
    note1 = Note(parent=meas1, onset=3, duration=1, pitch="F4")
    meas2 = Measure(parent=staff, onset=4)
    assert staff.duration == 4
    assert meas2.offset == 8
    staff.inherit_duration()
    assert staff.duration == 8
    note2 = Note(parent=meas2, onset=4, duration=0.125, pitch="F4")
    note1.tie = note2

    # case 4: short G tied to long G tied to short G
    note1 = Note(parent=meas2, onset=4.875, duration=0.125, pitch="G4")
    note2 = Note(parent=meas2, onset=5, duration=1, pitch="G4")
    note3 = Note(parent=meas2, onset=6, duration=0.125, pitch="G4")
    note1.tie = note2
    note2.tie = note3

    # case 5: two short tied A's that quantize to eighth (0.5), but each
    #     alone quantizes to zero
    note1 = Note(parent=meas2, onset=7, duration=0.24, pitch="A4")
    note2 = Note(parent=meas2, onset=7.24, duration=0.24, pitch="A4")
    note1.tie = note2

    # case 6: a single short B that quantizes to zero duration
    Note(parent=meas2, onset=7.5, duration=0.125, pitch="B4")

    print("Before quantize:")
    score.show(4)
    assert score.has_ties()
    score.quantize(2)
    print("After quantize:")
    score.show(4)
    assert not score.has_ties()

    # Check Score structure
    assert isinstance(score, Score)
    assert score.onset == 0.0
    assert score.duration == 4.0

    # Check Part
    assert len(score.content) == 1
    part = score.content[0]
    assert isinstance(part, Part)
    assert part.onset == 0.0

    # Check Staff
    assert len(part.content) == 1
    staff = part.content[0]
    assert isinstance(staff, Staff)
    assert staff.onset == 0.0

    # Check Measures
    assert len(staff.content) == 2
    m1, m2 = staff.content
    assert isinstance(m1, Measure)
    assert isinstance(m2, Measure)
    assert m1.onset == 0.0
    assert m1.duration == 4.0
    assert m2.onset == 4.0
    assert m2.duration == 4.0

    # Check Notes in Measure 1
    n1, n2, n3 = m1.content
    assert (
        n1.onset == 0.0 and n1.duration == 1.0 and cast(Note, n1).key_num == 60
    )
    assert (
        n2.onset == 2.0 and n2.duration == 1.0 and cast(Note, n2).key_num == 62
    )
    assert (
        n3.onset == 3.0 and n3.duration == 1.0 and cast(Note, n3).key_num == 65
    )

    # Check Notes in Measure 2
    n4, n5, n6 = m2.content
    assert (
        n4.onset == 5.0 and n4.duration == 1.0 and cast(Note, n4).key_num == 67
    )
    assert (
        n5.onset == 7.0 and n5.duration == 0.5 and cast(Note, n5).key_num == 69
    )
    assert (
        n6.onset == 7.5 and n6.duration == 0.5 and cast(Note, n6).key_num == 71
    )

    # Since we have a full score here, test some other core functions:
    # staff test:
    assert n1.staff == staff
    assert n5.staff == staff

    # measure test:
    # n1 is in measure 1
    assert n1.measure == m1
    # n5 is in measure 2
    assert n5.measure == m2

    # part test:
    assert n1.part == part
    assert n6.part == part

    # score test:
    # n3 is in score
    assert n3.score == score
    # n4 is in score
    assert n4.score == score


def test_lyrics():
    note1 = Note(onset=0, duration=1, pitch="C4", lyric="Hello")
    assert (
        str(note1) == "Note(onset=0.000, duration=1.000,"
        " lyric=Hello, pitch=C4/60)"
    )


def test_step():
    note1 = Note(onset=0, duration=1, pitch="C4")
    assert note1.step == "C"
    note2 = Note(onset=0, duration=1, pitch="D#4")
    assert note2.step == "D"
    note3 = Note(onset=0, duration=1, pitch="Fb4")
    assert note3.step == "F"
    note4 = Note(onset=0, duration=1, pitch="G##4")
    assert note4.step == "G"


def test_name():
    note1 = Note(onset=0, duration=1, pitch="C4")
    assert note1.name == "C"
    note2 = Note(onset=0, duration=1, pitch="D#4")
    assert note2.name == "D#"
    note3 = Note(onset=0, duration=1, pitch="Fb4")
    assert note3.name == "Fb"
    note4 = Note(onset=0, duration=1, pitch="G##4")
    assert note4.name == "G##"


def test_pitch_class():
    note1 = Note(onset=0, duration=1, pitch="C4")
    assert note1.pitch_class == 0
    note2 = Note(onset=0, duration=1, pitch="D#4")
    assert note2.pitch_class == 3
    note3 = Note(onset=0, duration=1, pitch="Fb4")
    assert note3.pitch_class == 4
    note4 = Note(onset=0, duration=1, pitch="G##4")
    assert note4.pitch_class == 9


def test_pitch_class_setter():
    note1 = Note(onset=0, duration=1, pitch="C4")
    note1.pitch_class = 4  # Set to E
    assert note1.pitch_class == 4
    assert note1.name_with_octave == "E4"

    note2 = Note(onset=0, duration=1, pitch="D#4")
    note2.pitch_class = 7  # Set to G
    assert note2.pitch_class == 7
    assert note2.name_with_octave == "G4"


def test_octave():
    note1 = Note(onset=0, duration=1, pitch="C4")
    assert note1.octave == 4
    note2 = Note(onset=0, duration=1, pitch="D#5")
    assert note2.octave == 5
    note3 = Note(onset=0, duration=1, pitch="Fb3")
    assert note3.octave == 3
    note4 = Note(onset=0, duration=1, pitch="G##2")
    assert note4.octave == 2


def test_octave_setter():
    note1 = Note(onset=0, duration=1, pitch="C4")
    note1.octave = 5
    assert note1.octave == 5
    assert note1.name_with_octave == "C5"

    note2 = Note(onset=0, duration=1, pitch="D#4")
    note2.octave = 3
    assert note2.octave == 3
    assert note2.name_with_octave == "D#3"


def test_key_num():
    note1 = Note(onset=0, duration=1, pitch="C4")
    assert note1.key_num == 60
    note2 = Note(onset=0, duration=1, pitch="D#5")
    assert note2.key_num == 75
    note3 = Note(onset=0, duration=1, pitch="Fb3")
    assert note3.key_num == 52
    note4 = Note(onset=0, duration=1, pitch="G##2")
    assert note4.key_num == 45
    try:
        note5 = Note(onset=0, duration=1, pitch=None)
        _ = note5.key_num
    except ValueError:
        pass
    else:
        assert False, "Expected ValueError"


def test_enharmonic():
    note1 = Note(onset=0, duration=1, pitch="C4")
    enharmonic1 = note1.enharmonic()
    assert enharmonic1.name_with_octave == "B#3"

    note2 = Note(onset=0, duration=1, pitch="D#4")
    enharmonic2 = note2.enharmonic()
    assert enharmonic2.name_with_octave == "Eb4"

    note3 = Note(onset=0, duration=1, pitch="Fb4")
    enharmonic3 = note3.enharmonic()
    assert enharmonic3.name_with_octave == "E4"

    note4 = Note(onset=0, duration=1, pitch="G##4")
    enharmonic4 = note4.enharmonic()
    assert enharmonic4.name_with_octave == "A4"

    note5 = Note(onset=0, duration=1, pitch=None)
    try:
        _ = note5.enharmonic()
    except ValueError:
        pass
    else:
        assert False, "Expected ValueError"


def test_upper_enharmonic():
    note1 = Note(onset=0, duration=1, pitch="C4")
    enharmonic1 = note1.upper_enharmonic()
    assert enharmonic1.name_with_octave == "Dbb4"

    note2 = Note(onset=0, duration=1, pitch="D#4")
    enharmonic2 = note2.upper_enharmonic()
    assert enharmonic2.name_with_octave == "Eb4"

    note3 = Note(onset=0, duration=1, pitch="Fb4")
    enharmonic3 = note3.upper_enharmonic()
    assert enharmonic3.name_with_octave == "Gbbb4"

    note4 = Note(onset=0, duration=1, pitch="G##4")
    enharmonic4 = note4.upper_enharmonic()
    assert enharmonic4.name_with_octave == "A4"

    note5 = Note(onset=0, duration=1, pitch=None)
    try:
        _ = note5.upper_enharmonic()
    except ValueError:
        pass
    else:
        assert False, "Expected ValueError"


def test_lower_enharmonic():
    note1 = Note(onset=0, duration=1, pitch="C4")
    enharmonic1 = note1.lower_enharmonic()
    assert enharmonic1.name_with_octave == "B#3"

    note2 = Note(onset=0, duration=1, pitch="D#4")
    enharmonic2 = note2.lower_enharmonic()
    assert enharmonic2.name_with_octave == "C###4"

    note3 = Note(onset=0, duration=1, pitch="Fb4")
    enharmonic3 = note3.lower_enharmonic()
    assert enharmonic3.name_with_octave == "E4"

    note4 = Note(onset=0, duration=1, pitch="G##4")
    enharmonic4 = note4.lower_enharmonic()
    assert enharmonic4.name_with_octave == "F####4"

    note5 = Note(onset=0, duration=1, pitch=None)
    try:
        _ = note5.lower_enharmonic()
    except ValueError:
        pass
    else:
        assert False, "Expected ValueError"


def test_simplest_enharmonic():
    note1 = Note(onset=0, duration=1, pitch="C4")
    enharmonic1 = note1.simplest_enharmonic()
    assert enharmonic1.name_with_octave == "C4"

    note2 = Note(onset=0, duration=1, pitch="D#4")
    enharmonic2 = note2.simplest_enharmonic()
    assert enharmonic2.name_with_octave == "Eb4"

    note3 = Note(onset=0, duration=1, pitch="Fb4")
    enharmonic3 = note3.simplest_enharmonic()
    assert enharmonic3.name_with_octave == "E4"

    note4 = Note(onset=0, duration=1, pitch="G##4")
    enharmonic4 = note4.simplest_enharmonic()
    assert enharmonic4.name_with_octave == "A4"

    note5 = Note(onset=0, duration=1, pitch=None)
    try:
        _ = note5.simplest_enharmonic()
    except ValueError:
        pass
    else:
        assert False, "Expected ValueError"


def test_clef():
    clef = Clef(onset=4.0, clef="treble")
    assert clef.clef == "treble"
    assert clef.onset == 4.0
    assert clef.duration == 0.0
    assert str(clef) == "Clef(onset=4.000, treble)"

    try:
        _ = Clef(onset=0, clef="wrong")
    except ValueError:
        pass
    else:
        assert False, "Expected ValueError for invalid clef"


def test_key_sig():
    key_sig = KeySignature(onset=4, key_sig=2)
    assert key_sig.key_sig == 2
    assert key_sig.onset == 4.0
    assert key_sig.duration == 0.0
    assert str(key_sig) == "KeySignature(onset=4.000, 2 sharps)"


def test_eventgroup():
    note1 = Note(onset=0, duration=1, pitch="C4")
    note2 = Note(onset=1, duration=1, pitch="E4")
    eg = Part(note1, note2, onset=0)
    try:
        _ = Part(note1, note2)
    except Exception:
        pass
    else:
        assert False, "Expected ValueError adding event with parent"
    assert eg.onset == 0.0
    assert eg.duration == 2.0
    assert len(eg.content) == 2
    assert str(eg) == "Part(onset=0.000, duration=2.000)"

    eg2 = Measure()
    try:
        _ = eg2.onset
    except ValueError as e:
        assert str(e) == "Onset time is not set"
    else:
        assert False, "Expected ValueError when onset is None"

    eg3 = Measure(duration=2)
    note1 = note1.insert_copy_into(eg3)
    note2 = note2.insert_copy_into(eg3)
    eg3.onset = 4
    assert eg3.onset == 4.0
    assert eg3.duration == 2.0
    assert len(eg3.content) == 2
    assert note1.onset == 4.0
    assert note2.onset == 5.0
    assert str(eg3) == "Measure(onset=4.000, duration=2.000)"

    part = Part(eg3, onset=0)
    assert part.onset == 0.0
    assert part.duration == 6.0
    assert len(part.content) == 1
    assert str(part) == "Part(onset=0.000, duration=6.000)"
    assert part.has_instanceof(Note)
    assert part.has_instanceof(Measure)
    assert not part.has_instanceof(Staff)
    assert part.has_rests() is False
    _ = Measure(Rest(onset=0, duration=4), onset=0, parent=part)
    assert part.has_rests() is True

    assert part.has_measures() is True

    assert part.has_chords() is False
    chord = Chord(
        Note(duration=1, pitch="D4"),
        Note(duration=1, pitch="F#4"),
        onset=6,
        parent=eg3,
    )
    assert part.has_chords() is True

    no_rests = part.remove_rests()
    assert no_rests.has_rests() is False
    buf = io.StringIO()
    no_rests.show(file=buf)
    correct = """Part(onset=0.000, duration=6.000)
    Measure(onset=0.000, duration=4.000)
    Measure(onset=4.000, duration=2.000)
        Note(onset=4.000, duration=1.000, pitch=C4/60)
        Note(onset=5.000, duration=1.000, pitch=E4/64)
        Chord(onset=6.000, duration=1.000)
            Note(onset=6.000, duration=1.000, pitch=D4/62)
            Note(onset=6.000, duration=1.000, pitch=F#4/66)
"""
    assert buf.getvalue() == correct
    assert chord.onset == 6.0
    assert chord.duration == 1.0
    assert str(chord) == "Chord(onset=6.000, duration=1.000)"

    eg4 = Staff(
        Note(pitch="C4"),
        Note(pitch="E4"),
        Note(pitch="G4"),
        Chord(Note(pitch="B3"), Note(pitch="D4")),
    )
    eg4.pack()
    eg4.show()
    assert eg4.onset == 0.0
    assert eg4.duration == 4.0
    assert len(eg4.content) == 4

    # let's test is_well_formed_full_part by moving the Part
    # content to a Staff and putting the Staff in the Part
    # and putting the Part in a Score - maybe it would have
    # been better to construct from scratch because it's a
    # pain to move all the members to a new parent.
    content = part.content
    part.content = []
    for elem in content:
        elem.parent = None  # ready to move to Staff
    Staff(*content, onset=0, parent=part)
    score = Score(part)
    assert score.is_well_formed_full_score() is True
    assert score.units_are_quarters is True
    assert score.units_are_seconds is False

    score.convert_to_seconds()
    assert score.units_are_quarters is False
    assert score.units_are_seconds is True

    # idempotent:
    score.convert_to_seconds()
    assert score.units_are_quarters is False
    assert score.units_are_seconds is True

    score.convert_to_quarters()
    score.convert_to_quarters()  # exercise a no-op path
    assert score.units_are_quarters is True
    assert score.units_are_seconds is False
    assert score.units_are_seconds is False  # exercise a no-op path
    assert score.is_well_formed_full_score() is True


def test_event_str():
    note = Note(onset=1, duration=2, pitch="A4")
    s = str(note)
    assert s.startswith("Note(")
    assert "onset=1.000" in s
    assert "duration=2.000" in s
    assert "pitch=A4" in s


def test_quantize_alignment():
    """Quantize should snap onsets and durations to multiples of 1/divisions."""
    # create a score with non-aligned onsets/durations
    score = Score.from_melody(
        pitches=[60, 62, 64],
        onsets=[0.13, 1.37, 2.88],
        durations=[0.47, 0.99, 0.26],
    )

    divisions = 4  # quarter divided into 4 -> quantum = 0.25
    for p in score.content:
        part = cast(Part, p)
        for event in part.content:
            assert event.units_are_quarters is True
            assert event.units_are_seconds is False
    score.convert_to_seconds()  # exercise conversion
    for p in score.content:
        part = cast(Part, p)
        for event in part.content:
            assert event.units_are_quarters is False
            assert event.units_are_seconds is True
    score.convert_to_quarters()  # back to quarters
    score.show()

    score.quantize(divisions)

    # traverse notes and assert onsets and durations are multiples of 1/divisions
    for part in score.content:
        for staff in getattr(part, "content", []):
            for measure in getattr(staff, "content", []):
                for note in getattr(measure, "content", []):
                    # skip non-Note objects
                    if not isinstance(note, Note):
                        continue
                    on_mul = note.onset * divisions
                    dur_mul = note.duration * divisions
                    assert math.isclose(
                        on_mul, round(on_mul), abs_tol=1e-8
                    ), f"onset {note.onset} not on grid for divisions={divisions}"
                    assert math.isclose(
                        dur_mul, round(dur_mul), abs_tol=1e-8
                    ), f"duration {note.duration} not on grid for divisions={divisions}"


def test_merge_tied_notes():
    """Test merging tied notes into single notes."""
    score = Score.from_melody(
        pitches=["C4", "C4", "E4", "E4", "E4", "G4"],
        durations=[1.0, 0.5, 0.5, 0.5, 1.0, 1.0],
        ties=[True, False, True, True, False, False],
    )
    print("Before merging tied notes:")
    score.show(4)

    score = score.merge_tied_notes()

    print("After merging tied notes:")
    score.show(4)

    # Check that tied notes have been merged correctly
    part: Part = score.content[0]  # type: ignore (score.content[i] is a Part)
    notes = [event for event in part.content if isinstance(event, Note)]
    assert len(notes) == 3  # Should be 3 notes after merging

    assert (
        notes[0].pitch
        and notes[0].pitch.key_num == 60
        and notes[0].duration == 1.5
    )
    assert (
        notes[1].pitch
        and notes[1].pitch.key_num == 64
        and notes[1].duration == 2.0
    )
    assert (
        notes[2].pitch
        and notes[2].pitch.key_num == 67
        and notes[2].duration == 1.0
    )


def test_pack():
    """Test merging tied notes into single notes."""
    score = Score.from_melody(
        pitches=["C4", "D4", "E4", "F4", "G4", "A4"],
        onsets=[0, 2, 3, 4, 5, 6],
        durations=[1.0, 0.5, 0.5, 0.5, 1.0, 1.0],
    )
    print("Before packing:")
    score.show(4)
    score.pack(0.0, sequential=True)
    print("After packing:")
    score.show(4)
    assert score.ismonophonic()
    assert score.parts_are_monophonic()


def test_well_formed():
    # Test is_well_formed_full_score method for various scenarios
    score = Score(
        Part(
            Staff(
                Measure(
                    Note(onset=0, duration=1.0, pitch="C4"),
                    Note(onset=1.0, duration=1.0, pitch="E4"),
                    Chord(
                        Note(onset=2.0, duration=1.0, pitch="G4"),
                        Note(onset=2.0, duration=1.0, pitch="B4"),
                        onset=2.0,
                    ),
                )
            )
        )
    )
    score1: Score = score.copy()  # type: ignore

    assert score1.is_well_formed_full_score() is True
    chord: Chord = next(score1.find_all(Chord))  # type: ignore
    Chord(
        Note(onset=2, duration=1.0, pitch="D5"),
        Note(onset=2, duration=1.0, pitch="F#5"),
        onset=2,
        parent=chord,
    )
    print("Modified score with invalid chord:")
    score1.show(4)
    assert score1.is_well_formed_full_score() is False

    score1: Score = score.copy()  # type: ignore
    assert score1.is_well_formed_full_score() is True
    measure: Measure = next(score1.find_all(Measure))  # type: ignore
    Measure(
        Note(onset=3.0, duration=1.0, pitch="A4"),
        Note(onset=4.0, duration=1.0, pitch="C5"),
        onset=3.0,
        parent=measure,
    )
    assert score1.is_well_formed_full_score() is False

    # insert a Staff directly into Score
    score1: Score = score.copy()  # type: ignore
    assert score1.is_well_formed_full_score() is True
    Staff(
        Note(onset=0, duration=1.0, pitch="D4"),
        Note(onset=1.0, duration=1.0, pitch="F#4"),
        onset=0,
        parent=score1,
    )
    assert score1.is_well_formed_full_score() is False

    # insert a Note directly into Part
    score1: Score = score.copy()  # type: ignore
    assert score1.is_well_formed_full_score() is True
    part: Part = score1.content[0]  # type: ignore
    Note(onset=0, duration=1.0, pitch="D4", parent=part)
    assert score1.is_well_formed_full_score() is False

    # put a rest in a staff
    score1: Score = score.copy()  # type: ignore
    assert score1.is_well_formed_full_score() is True
    staff: Staff = next(score1.find_all(Staff))  # type: ignore
    Rest(onset=2.0, duration=1.0, parent=staff)
    assert score1.is_well_formed_full_score() is False


def make_a_full_score() -> Score:
    """Helper function to create a well-formed full score for testing."""
    score = Score(
        Part(
            Staff(
                Measure(
                    Note(onset=0, duration=1.0, pitch="C4"),
                    Note(onset=1.0, duration=1.0, pitch="E4"),
                ),
                number=1,
            ),
            Staff(
                Measure(
                    Note(onset=0, duration=1.0, pitch="G4"),
                    Note(onset=1.0, duration=1.0, pitch="B4"),
                    Note(onset=2.0, duration=1.0, pitch="D5"),
                ),
                number=2,
            ),
            # duration None tests a case in the Part constructor:
            instrument="Flute",
            duration=None,
        ),
        Part(
            Staff(
                Measure(
                    Note(onset=0, duration=1.0, pitch="D4"),
                    Note(onset=1.0, duration=1.0, pitch="F#4"),
                ),
                number=3,
            ),
            instrument="Violin",
        ),
    )
    return score


def test_collapse_parts():
    """Test collapsing parts in a score into a single part
    where staff is specified."""
    score = make_a_full_score()
    score2: Score = score.copy()  # type: ignore
    print("Original score with multiple parts:")
    score.show(4)
    collapsed_score = score.collapse_parts(part=[0], staff=2)

    print("Collapsed score with single part:")
    collapsed_score.show(4)

    # Check that the collapsed score has only one part
    assert len(collapsed_score.content) == 1
    part: Part = collapsed_score.content[0]  # type: ignore
    # Check that the part contains all notes from staff 2
    notes: List[Note] = list(part.find_all(Note))  # type: ignore
    assert len(notes) == 3

    expected_pitches = {67, 71, 74}  # G4, B4, D5
    actual_pitches = {note.pitch.key_num for note in notes if note.pitch}
    assert actual_pitches == expected_pitches

    score: Score = score2.copy()  # type: ignore  # with has_ties=False:
    collapsed_score = score.collapse_parts(part=[0], staff=2, has_ties=False)

    print("Collapsed score with single part:")
    collapsed_score.show(4)

    # Check that the collapsed score has only one part
    assert len(collapsed_score.content) == 1
    part: Part = collapsed_score.content[0]  # type: ignore
    # Check that the part contains all notes from staff 2
    notes: List[Note] = list(part.find_all(Note))  # type: ignore
    assert len(notes) == 3

    expected_pitches = {67, 71, 74}  # G4, B4, D5
    actual_pitches = {note.pitch.key_num for note in notes if note.pitch}
    assert actual_pitches == expected_pitches

    # is_flat test
    assert score.is_flat

    # flatten test
    score = score2.flatten(collapse=True)
    # expect part instrument to be None
    part: Part = score.content[0]  # type: ignore
    assert part.instrument is None

    # change instruments to match and try again
    for p in score2.content:
        part = cast(Part, p)
        part.instrument = "Piano"
    score = score2.flatten(collapse=True)
    # expect part instrument to be Piano
    part: Part = score.content[0]  # type: ignore
    assert part.instrument == "Piano"

    # more is_flat testing, add Note to Score directly
    # start with score, which is properly flattened
    score3: Score = score.copy()  # type: ignore
    Note(onset=0, duration=1.0, pitch="C4", parent=score3)
    assert not score3.is_flat()

    # # more is_flat testing, add Note to Part directly
    # score3 : Score = score.copy()  # type: ignore
    # part : Part = score3.content[0]  # type: ignore
    # Note(onset=0, duration=1.0, pitch="C4", parent=part)
    # assert not score3.is_flat()

    # more is_flat testing, add non-note to a Part directly
    score3: Score = score.copy()  # type: ignore
    part: Part = score3.content[0]  # type: ignore
    Rest(onset=3, duration=1.0, parent=part)
    assert not score3.is_flat()

    # more is_flat testing, make a tied note
    score3: Score = score.copy()  # type: ignore
    part: Part = score3.content[0]  # type: ignore
    # first note is C4 at onset 0, duration 1.0
    note = Note(onset=1, duration=1.0, pitch="C4", parent=part)
    first_note = next(part.find_all(Note))  # type: ignore
    first_note.tie = note
    print("Score after adding tied Note in Part:")
    score3.show(4)
    assert not score3.is_flat()


def test_note_containers():
    """Test that Note containers work as expected."""
    score = make_a_full_score()
    containers = score.note_containers()
    print("note containers:", containers)

    # Expecting 3 containers
    assert len(containers) == 3
    assert containers[0] == score.content[0].content[0]  # type: ignore
    assert containers[1] == score.content[0].content[1]  # type: ignore
    assert containers[2] == score.content[1].content[0]  # type: ignore

    # test case where a part has no staff and is a container
    # remove staff from 2nd part
    part: Part = score.content[1]  # type: ignore
    part.flatten(in_place=True)  # move content up to Part level
    containers = score.note_containers()
    print("note containers after flattening part 2:", containers)
    # Expecting 3 containers now
    assert len(containers) == 3
    assert containers[0] == score.content[0].content[0]  # type: ignore
    assert containers[1] == score.content[0].content[1]  # type: ignore
    assert containers[2] == score.content[1]  # type: ignore


def test_remove_measures():
    """Test removing measures from a score."""
    score = make_a_full_score()
    # add a spurious top-level note to test that it is preserved
    Note(onset=4, duration=1.0, pitch="C5", parent=score)
    # add a spurious note in a staff to test that it is preserved
    part: Part = score.content[0]  # type: ignore
    staff: Staff = part.content[0]  # type: ignore
    Note(onset=8, duration=1.0, pitch="E5", parent=staff)
    print("Original score:")
    score.show(4)

    # Remove measures 1 from part 0
    modified_score = score.remove_measures()

    print("Modified score after removing measures:")
    modified_score.show(4)

    # Check that the measures have been removed correctly
    part: Part = modified_score.content[0]  # type: ignore
    assert len(part.content[0].content) == 3  # type: ignore
    assert len(part.content[1].content) == 3  # type: ignore

    part: Part = modified_score.content[1]  # type: ignore
    assert len(part.content[0].content) == 2  # type: ignore

    # Check that the top-level note is still present
    top_level_notes = [
        event for event in modified_score.content if isinstance(event, Note)
    ]
    assert len(top_level_notes) == 1
    assert (
        top_level_notes[0].pitch and top_level_notes[0].pitch.key_num == 72
    )  # C5

    # Check that the spurious note in staff is still present
    part: Part = modified_score.content[0]  # type: ignore
    staff: Staff = part.content[0]  # type: ignore
    spurious_notes = [
        event
        for event in staff.content
        if isinstance(event, Note) and event.onset == 8
    ]
    assert len(spurious_notes) == 1
    assert (
        spurious_notes[0].pitch and spurious_notes[0].pitch.key_num == 76
    )  # E5


def test_shift_content():
    """Test to see if chord content is shifted when chords are in
    sequence, and also test shifting content in a score."""
    score = Score(
        Part(
            Staff(
                Measure(
                    Chord(Note(pitch="C4"), Note(pitch="E4")),
                    Chord(Note(pitch="D4"), Note(pitch="F4")),
                )
            )
        )
    )
    print("Original score:")
    score.show(4)
    # check times in chords
    part = score.content[0]  # type: ignore
    staff = part.content[0]  # type: ignore
    measure = staff.content[0]  # type: ignore
    chord1 = measure.content[0]  # type: ignore
    chord2 = measure.content[1]  # type: ignore
    assert chord1.onset == 0.0
    assert chord2.onset == 1.0
    note1 = chord1.content[0]  # type: ignore
    note2 = chord1.content[1]  # type: ignore
    note3 = chord2.content[0]  # type: ignore
    note4 = chord2.content[1]  # type: ignore
    assert note1.onset == 0.0
    assert note2.onset == 0.0
    assert note3.onset == 1.0
    assert note4.onset == 1.0

    # Shift content by 2 quarters
    score.time_shift(2.0, content_only=True)

    print("Modified score after shifting content by 2 quarters:")
    score.show(4)

    # Check that all notes have been shifted correctly
    assert score.onset == 0.0
    part = score.content[0]  # type: ignore
    assert part.onset == 2.0
    staff = part.content[0]  # type: ignore
    assert staff.onset == 2.0
    measure = staff.content[0]  # type: ignore
    assert measure.onset == 2.0
    chord1 = measure.content[0]  # type: ignore
    chord2 = measure.content[1]  # type: ignore
    assert chord1.onset == 2.0
    assert chord2.onset == 3.0
    note1 = chord1.content[0]  # type: ignore
    note2 = chord1.content[1]  # type: ignore
    note3 = chord2.content[0]  # type: ignore
    note4 = chord2.content[1]  # type: ignore
    assert note1.onset == 2.0
    assert note2.onset == 2.0
    assert note3.onset == 3.0
    assert note4.onset == 3.0
    assert note4.measure is measure


def test_parts_are_monophonic():
    """Test parts_are_monophonic method."""
    # make a score with two monophonic parts
    score = Score(
        Part(
            Staff(
                Measure(
                    Note(onset=0.0, duration=1.0, pitch="C4"),
                    Note(onset=1.0, duration=1.0, pitch="D4"),
                )
            ),
            instrument="Flute",
        ),
        Part(
            Staff(
                Measure(
                    Note(onset=0.0, duration=1.0, pitch="E4"),
                    Note(onset=1.0, duration=1.0, pitch="F4"),
                )
            ),
            instrument="Oboe",
        ),
    )
    assert score.parts_are_monophonic() is True
    assert score.ismonophonic() is False

    # Add a chord to make it non-monophonic
    part: Part = score.content[0]  # type: ignore
    staff: Staff = part.content[0]  # type: ignore
    measure: Measure = staff.content[0]  # type: ignore
    Chord(
        Note(onset=5.0, duration=1.0, pitch="A4"),
        Note(onset=5.0, duration=1.0, pitch="C5"),
        onset=5.0,
        parent=measure,
    )
    assert score.parts_are_monophonic() is False



"""Sliding window algorithm.

The algorithm segments a musical score into fixed-size windows that can optionally
overlap. Each window contains all notes that are sounding within its time boundaries.

Author
------
Peter Harrison
"""

from typing import Iterable, Iterator, List, Optional, Union, cast

from amads.algorithms.slice.slice import Slice
from amads.core.basics import Note, Score
from amads.core.utils import float_range


class Window(Slice):
    """A fixed-size window of a musical score. candidate_notes that overlap
    with this interval are copied and clipped to fit within the window.
    Notes that overlap less than 1.0e-6 duration units (whether beats or
    seconds) are mostly excluded from the window to reduce numerical issues.
    An exception is made for notes that are so short that they do not overlap
    any window by at least 1.0e-6 duration units. (This seems far-fetched,
    but zero-length notes representing grace notes are one possibility to
    consider; there may be others.)

    Additional details that you may not need: For very short notes, the
    window is considered closed on the left and open on the right, so
    that the window is considered to contain the note if it starts at
    the same time as the window, and a note is not in the window if it
    starts at the offset time of the window. To guarantee that
    zero-length notes are included in only one window, the offset of a
    window should be identical to the onset of the next window.
    sliding_window() takes care of this by default, but if the Window
    constructor is used directly where arithmetic with time and size is
    inexact, this may not be the case.

    <small>**Author**: Peter Harrison</small>

    Parameters
    ----------
    time : float
        The reference time for this window
    size : float
        The size of the window in time units
    align : str
        How to align the window relative to the reference time:
        "left": window starts at reference time, "center": reference time
        is at window center, or "right": window ends at reference time.
    candidate_notes : Iterable[Note]
        Notes to consider for inclusion in this window, sorted by onset
        time and pitch
    skip : int, default=0
        Index to start searching from in candidate_notes. This is used
        to optimize performance when iterating through multiple
        windows - each window can tell the next window which notes
        it can safely skip because they end before the window starts.
    """

    def __init__(
        self,
        time: float,
        size: float,
        align: str,
        candidate_notes: Iterable[Note],
        skip: int = 0,
    ):
        match align:
            case "left":
                onset = time
            case "center":
                onset = time - size / 2
            case "right":
                onset = time - size
            case _:
                raise ValueError(f"Invalid value passed to `align`: {align}")

        offset = onset + size

        super().__init__(
            original_notes=[],
            onset=onset,
            duration=size,
        )

        self.time = time
        self.align = align
        # skip logic: skip is the lowest index of candidate_notes for which
        # candidate_notes[skip].offset >= onset. The next window can start
        # searching from this index.

        candidate_notes = list(candidate_notes)
        self.skip = len(candidate_notes)

        for i in range(skip, len(candidate_notes)):
            note = candidate_notes[i]

            if note.offset < onset:
                # The note finished before the window starts.
                continue
            else:  # note overlaps window, start no later than this
                # when searching for overlaps with the next window
                self.skip = min(self.skip, i)

            if note.onset >= offset:
                # The note starts after the window finishes.
                # All the remaining notes in candidate_notes will have even later onsets,
                # so we don't need to check them for this window.
                # They might be caught by future windows though.
                break

            # note.onset < offset, so it starts in window
            if note.offset - offset < 1.0e-6:
                # The note will not overlap with the next window.
                # Since it starts in this window, include it.
                pass
            elif offset - note.onset < 1.0e-6:
                # after clipping to window boundaries, this note
                # will be shorter than 1.0e-6, but it extends into
                # the next window, so ignore the small overlap here
                continue

            self.original_notes.append(note)

            # We use deepcopy_into instead of creating a new Note because we want to
            # preserve any other attributes that might be useful in downstream tasks.
            note = note.insert_copy_into(parent=self)
            # Clip the note to fit within the window
            note.onset = max(note.onset, onset)
            note.offset = min(note.offset, offset)


def sliding_window(
    passage: Union[Score, Iterable[Note]],
    size: float,
    step: float = 1.0,
    align: str = "right",
    onset: float = 0.0,
    offset: Optional[float] = None,
    times: Optional[Iterable[float]] = None,
) -> Iterator[Window]:
    """Slice a score into (possibly overlapping) windows of a given size.

    <small>**Author**: Peter Harrison</small>

    Parameters
    ----------
    passage : Score or Iterable[Note]
        The musical passage to be windowed
    size : float
        The size (duration) of each window (time units)
    step : float, default=1.0
        The step size to take between windows (time units).
        For example, if step is 0.1, then a given slice will start 0.1 time units
        after the previous slice started. Note that if step is smaller than size,
        successive windows will overlap
    align : str, default="right"
        Each generated window has a `time` property that points to a
        particular timepoint in the musical passage. The `align` parameter determines
        how the window is aligned to this timepoint:
        - "left": the window starts at ``window.time``
        - "center": ``window.time`` corresponds to the midpoint of the window
        - "right": the window finishes at ``window.time``
    onset : float, default=0.0
        The desired time of the first window
    offset : float, optional
        If set, the windowing will stop once the offset time is reached.
        Following the behaviour of Python's built-in range function,
        ``offset`` is not treated inclusively, i.e. the last window will
        not include ``offset``. The returned iterator will stop early
        the last window is empty (i.e. contains no notes) and there are
        no more notes to process.
    times : Iterable[float], optional
        Optional iterable of times to generate windows for. If provided,
        `onset` and `offset` are ignored. The returned iterator will
        stop once all times have been processed or when an empty window
        is generated and there are no more notes to process.

    Returns
    -------
    Iterator[Window]
        Iterator over the windows
    """
    if isinstance(passage, Score):
        if not passage.is_flat_and_collapsed():
            raise NotImplementedError(
                "Currently this function only supports flat scores with a "
                "single Part. You can flatten a score using "
                "`score.flatten(collapse=True)`."
            )
        notes = passage.find_all(Note)
    else:
        notes = passage

    notes = cast(List[Note], list(notes))
    notes.sort(key=lambda n: (n.onset, n.pitch))

    # We could rely on Window to obey `align`, but here we convert onset and
    # offset to always use "left". By using left, we can guarantee when
    # step == size that each window time computed by repeated addition of
    # step will exactly equal each previous window offset, computed as
    # onset + size in Window. This avoids any floating point rounding error
    # that could affect the windowing. (This is not a problem in the typical
    # cases where window size is 1 or a power of 2 that will be computed
    # exactly due to binary representations of floats.)
    match align:
        case "left":
            pass
        case "center":
            onset -= size / 2
            if offset is not None:
                offset -= size / 2
        case "right":
            onset -= size
            if offset is not None:
                offset -= size
        case _:
            raise ValueError(f"Invalid value passed to `align`: {align}")

    if times is None:
        window_times = float_range(onset, offset, step)
    else:
        for par, default in [("onset", 0.0), ("offset", None), ("step", 1.0)]:
            provided = globals()[par]
            if provided != default:
                raise ValueError(
                    f"`{par}` was set to {provided} but `times` was also provided"
                )

        window_times = times

    skip = 0

    for time in window_times:
        window = Window(time, size, "left", notes, skip)

        yield window

        skip = window.skip

        if skip == len(notes):
            break

    skip = 0
    if times is not None:
        for par, default in [("onset", 0.0), ("offset", None), ("step", 1.0)]:
            provided = globals()[par]
            if provided != default:
                raise ValueError(
                    f"`{par}` was set to {provided} but `times` was also provided"
                )
        for time in times:
            window = Window(time, size, align, notes, skip)

            yield window

            skip = window.skip
            if skip == len(notes):
                break
    else:  # compute windows equally spaced by step
        match align:
            case "left":
                time = onset
            case "center":
                time = onset - size / 2
            case "right":
                time = onset - size
            case _:
                raise ValueError(f"Invalid value passed to `align`: {align}")
        while (offset is None) or (time < offset):
            window = Window(time, size, "left", notes, skip)

            yield window

            # if step == size, windows are back-to-back, so the next onset
            # is exactly equal to the previous offset. Avoid
            onset = window.offset if step == size else onset + step
            time += step
            skip = window.skip
            if skip == len(notes):
                break



"""
Time variability.

This module provides various functions for calculating rhythmic variability in music,
including the normalized pairwise variability index (nPVI) and variations thereof.

<small>**Author**: Huw Cheston (2025)</small>

References
----------

- Daniele, J. R., & Patel, A. D. (2013). An Empirical Study of
    Historical Patterns in Musical Rhythm. *Music Perception*, 31(1), 10-18.
      https://doi.org/10.1525/mp.2013.31.1.10

- Condit-Schultz, N. (2019). Deconstructing the nPVI: A Methodological
    Critique of the Normalized Pairwise Variability Index as Applied to
    Music. *Music Perception*,
    36(3), 300–313. https://doi.org/10.1525/mp.2019.36.3.300

- VanHandel, L., & Song, T. (2010). The role of meter in compositional style
    in 19th-century French and German art song. *Journal of New Music
    Research*, 39, 1–11. https://doi.org/10.1080/09298211003642498
"""

import math
from typing import Iterable

__author__ = "Huw Cheston"


TINY = 1e-4


def _validate_inputs(
    durations: Iterable[float], _kind: str = "durations", _min: float = 2
) -> None:
    """
    Validates an input list (usually durations) and raises errors if:
    - insufficient number of elements,
    - any element (duration) is zero or negative.
    """

    if len(durations) < _min:
        raise ValueError(
            f"Must have at least {_min} {_kind}, but got {len(durations)} duration(s)"
        )

    if any(d <= 0.0 for d in durations):
        raise ValueError(f"All {_kind} must be positive, but got {durations}!")


def _normalized_pairwise_calculation(a_ioi: float, c_ioi: float) -> float:
    """Normalized pairwise calculation for an antecedent and consequent IOI, as in Condit-Schultz (2019)."""
    return 200 * abs((a_ioi - c_ioi) / (a_ioi + c_ioi))


# flake8: noqa: W605
def normalized_pairwise_variability_index(
    durations: Iterable[float],
) -> float:
    r"""
    Calculates the normalised pairwise variability index (nPVI).

    The nPVI is a measure of variability between successive elements
    in a sequence. The equation is:

    $\text{nPVI} = \frac{100}{m-1} \times \sum\limits_{k=1}^{m-1}
    \left| \frac{d_k - d_{k+1}}{\frac{d_k + d_{k+1}}{2}} \right|$

    where $m$ is the number of intervals, and $d_k$ is the duration
    of the $k^{th}$ interval.

    A completely regular stream of equal durations returns 0 variability.
    High difference between successive items returns a high nPVI value
    irrespective of other (e.g., metrical) considerations.

    Parameters
    ----------
    durations : Iterable[float]
        the durations to analyse

    Returns
    -------
    float
        the extracted nPVI value.

    Examples
    -------

    A completely regular stream of equal durations returns 0 variability.

    >>> normalized_pairwise_variability_index([1., 1., 1., 1.])
    0.

    This example is from Daniele & Patel (2013, Appendix).

    >>> durs = [1., 1/2, 1/2, 1., 1/2, 1/2, 1/3, 1/3, 1/3, 2., 1/3, 1/3, 1/3, 1/3, 1/3, 1/3, 3/2, 1., 1/2]
    >>> x = normalized_pairwise_variability_index(durations=durs)
    >>> round(x, 1)
    42.2

    These next examples are from Condit-Schultz's 2019 critique (Figure 2).

    >>> normalized_pairwise_variability_index([0.25, 0.25, 0.25, 0.25, 1., 0.25, 0.25, 0.25, 0.25, 1.])
    40.

    >>> normalized_pairwise_variability_index([0.25, 0.25, 0.5, 1., 0.25, 0.25, 0.5, 1.])
    55.2

    >>> normalized_pairwise_variability_index([0.5, 0.25, 0.25, 1., 0.5, 0.25, 0.25, 1.])
    62.8

    >>> normalized_pairwise_variability_index([2, 1, 2, 1])
    66.66

    >>> normalized_pairwise_variability_index([0.5, 1, 2, 1, 0.5])
    66.66

    >>> normalized_pairwise_variability_index([2, 1, 0.5, 1, 0.5, 0.25])
    66.66
    """

    _validate_inputs(durations)
    numerator = (
        sum(
            [
                abs((k - k1) / ((k + k1) / 2))
                for (k, k1) in zip(durations, durations[1:])
            ]
        )
        * 100
    )
    denominator = len(durations) - 1
    return numerator / denominator


def normalized_pairwise_calculation(durations: Iterable[float]) -> list[float]:
    r"""
    Calculates the normalized pairwise calculation (nPC) for a list of
    durations, as defined by Condit-Schultz (2019).

    The nPVI is equivalent to the arithmetic mean of a set of nPC values.
    The equation for nPC can be written as:

    $\text{nPC} = 200 * \biggl{|} \frac{\text{antecedent IOI} -
                                        \text{consequent IOI}}
        {\text{antecedent IOI} + \text{consequent IOI}} \biggr{|}$

    Parameters
    ----------
    durations : Iterable[float]
        the durations to analyse

    Returns
    -------
    Iterable[float]
        the extracted nPC value(s) for every pair of
        antecedent-consequent durations

    Examples
    --------

    From Condit-Schultz (2019) figure 3.

    >>> normalized_pairwise_calculation([1, 15])
    [175.]

    >>> normalized_pairwise_calculation([1, 7])
    [150.]

    >>> normalized_pairwise_calculation([1, 5])
    [133.33]

    >>> normalized_pairwise_calculation([1, 4])
    [120.]

    >>> normalized_pairwise_calculation([1, 3])
    [100.]

    >>> normalized_pairwise_calculation([1, 2])
    [66.67]

    >>> normalized_pairwise_calculation([2, 3])
    [40.]

    >>> normalized_pairwise_calculation([1, 1])
    [0.]

    """

    _validate_inputs(durations)
    return [
        _normalized_pairwise_calculation(a_ioi, c_ioi)
        for a_ioi, c_ioi in zip(durations, durations[1:])
    ]


def isochrony_proportion(
    durations: Iterable[float],
) -> float:
    r"""
    Calculates the isochrony proportion (IsoP) for a list of durations, as defined in Condit-Schultz (2019):
    "by iterating over every pair of successive IOIs in a rhythm,
    counting the pairs that are identical,
    and dividing this count by the total number of pairs (one less than the total number of IOIs)."

    Condit-Schultz 2019 demonstrates that the IsoP accounts for a large degree of the variance in the nPVI.

    Parameters
    ----------
    durations (Iterable[float]): the durations to analyse

    Returns
    -------
    float: the extracted IsoP value

    Examples
    --------
    >>> isochrony_proportion([1, 1, 2, 2, 1, 0.5])
    0.4

    """

    _validate_inputs(durations)
    isochronous = 0
    # Iterate over every pair of successive IOIs in a rhythm
    for a1, a2 in zip(durations, durations[1:]):
        # If both IOIs are the same
        if math.isclose(a1, a2, abs_tol=TINY):
            # Increase the counter by one
            isochronous += 1
    # Divide the counter by the total number of pairs (one less than the total number of IOIs)
    denominator = len(durations) - 1
    return isochronous / denominator


def pairwise_anisochronous_contrast_index(
    durations: Iterable[float],
) -> float:
    r"""
    Calculates the pairwise anisochronous contrast index (pACI) for a list of durations.

    Defined in Condit-Schultz (2019), the pACI is equivalent to the nPVI,
    except that it "factors out" isochronous pairs of IOIs.
    This means that it is sensitive to changes in the frequencies of IOIs pairs such as 2:1, 3:1, without
    being "overwhelmed by isochronous pairs" (1:1).

    Parameters
    ----------
    durations (Iterable[float]): the durations to analyse

    Returns
    -------
    float: the extracted pACI value

    """

    _validate_inputs(durations)
    all_npcs = []
    # Iterate over successive IOIs
    for a1, a2 in zip(durations, durations[1:]):
        # If the two IOIs are NOT isochronous
        if not math.isclose(a1, a2, abs_tol=TINY):
            # Calculate the nPC for this pair of durations
            npc = _normalized_pairwise_calculation(a1, a2)
            all_npcs.append(npc)
    # Raise errors as required
    if len(all_npcs) == 0:
        raise ValueError(
            "No non-isochronous pairs were found, cannot calculate pACI."
        )
    # Return the average
    return sum(all_npcs) / len(all_npcs)


def phrase_normalized_pairwise_variability_index(
    durations: Iterable[float],
    phrase_boundaries: Iterable[float],
) -> float:
    r"""
    Calculates the phrase-normalized pairwise variability index (pnPVI), as defined by VanHandel & Song (2010).

    The pnNPVI calculates the nPVI, ignoring cases where pairs of IOIs straddle a phrase boundary.
    To do so, we also need a list of times for where those phrase boundaries fall.

    Parameters
    ----------
    durations (Iterable[float]): the durations to analyse
        phrase_boundaries (Iterable[float]): the phrase boundaries to analyse

    Returns
    -------
    float: the extracted pnNPVI value

    """

    # Validate all inputs
    _validate_inputs(durations)
    _validate_inputs(
        phrase_boundaries, _kind="phrase boundaries", _min=1
    )  # need at least one phrase boundary
    # We use the "counter" to keep track of where we are in the phrase
    counter = 0
    all_npcs = []
    # Iterate over pairs of IOIs
    for a1, a2 in zip(durations, durations[1:]):
        # This is where we "end" the current boundary: add both IOIs to the counter
        end_pos = counter + a1 + a2
        # If we're not straddling any of the phrase boundaries
        if not any([counter <= pb <= end_pos for pb in phrase_boundaries]):
            # Calculate the nPC
            npc = _normalized_pairwise_calculation(a1, a2)
            all_npcs.append(npc)
        # Add the first IOI to the counter to move forwards in time for the next pair
        counter += a1
    # Raise errors as required
    if len(all_npcs) == 0:
        raise ValueError(
            "All pairs of IOIs cross phrase boundaries, cannot calculate pnNPVI."
        )
    # Return the average of all nPC values
    return sum(all_npcs) / len(all_npcs)



"""
Export a Score as a standard MIDI file using PrettyMIDI library.

Functions
---------
- pretty_midi_midi_export(score: Score, filename: str) -> None:
    Exports a `Score` object as a MIDI file using PrettyMIDI.

Usage
-----
Do not use this module directly; see writescore.py.

Notes
-----
See `export_midi` notes for translation-to-MIDI-file details.

"""

from math import isclose
from typing import cast

import pretty_midi as pm

from amads.core.basics import EventGroup, KeySignature, Note, Part, Score, Staff

__author__ = "Roger B. Dannenberg"


def string_to_velocity(dynamic: str) -> int:
    """Convert a dynamic string (e.g., 'p', 'mf', 'f') to a MIDI velocity value."""
    dynamic_map = {
        "pppp": 10,
        "ppp": 23,
        "pp": 36,
        "p": 49,
        "mp": 62,
        "mf": 75,
        "f": 99,
        "ff": 102,
        "fff": 114,
        "ffff": 127,
    }
    return dynamic_map.get(dynamic.lower(), 100)  # Default to 100 if unknown


tied_to_notes = None  # global to help merge tied notes


def add_note_to_instrument(note: Note, instrument: pm.Instrument) -> None:
    """Add a Note to a PrettyMIDI Instrument."""
    global tied_to_notes

    if note in tied_to_notes:  # type: ignore
        return  # Note is tied from another, already handled
    dynamic = note.dynamic if note.dynamic is not None else 100
    if isinstance(dynamic, str):
        dynamic = string_to_velocity(dynamic)
    pm_note = pm.Note(
        velocity=dynamic,
        pitch=round(note.key_num),
        start=note.onset,
        end=note.onset + note.tied_duration,
    )
    if note.tie:
        note = note.tie
        while note:
            tied_to_notes[note] = True  # type: ignore
            note = note.tie  # type: ignore - follow chain of ties

    instrument.notes.append(pm_note)


def add_key_signature(
    onset: float, key_sig: int, key_signatures: list[pm.KeySignature]
) -> None:
    """Add a Key Signature change to the PrettyMIDI time signatures list."""
    # PrettyMIDI uses key number from -7 (7 flats) to +7 (7 sharps)
    location = len(key_signatures)  # default to appending at end
    for index, ts in enumerate(key_signatures):
        if isclose(ts.time, onset, abs_tol=1e-3) and ts.key_number == key_sig:
            return  # Key signature already exists at this time
        elif ts.time > onset + 1e-3:
            location = index
            break
    # now location is where to insert a new key signature
    pm_key_sig = pm.KeySignature(key_number=key_sig, time=onset)
    key_signatures.insert(location, pm_key_sig)


def add_eventgroup_to_instrument(
    evgroup: EventGroup,
    instrument: pm.Instrument,
    key_signatures: list[pm.KeySignature],
) -> None:
    """Add events from a Part or Staff to a PrettyMIDI Instrument.

    This function is recursive to process Notes in Measures, Chords, etc.
    """
    for event in evgroup.content:
        if isinstance(event, KeySignature):
            add_key_signature(event.onset, event.key_sig, key_signatures)
        elif isinstance(event, Note):
            note = cast(Note, event)
            add_note_to_instrument(note, instrument)
        elif isinstance(event, EventGroup):
            add_eventgroup_to_instrument(event, instrument, key_signatures)
        # else event is unhandled, hopefully a Rest, or Clef


def pretty_midi_export(
    score: Score,
    filename: str,
    format: str,
    show: bool = False,
    display: bool = False,
) -> None:
    """
    Export a Score as a standard MIDI file using PrettyMIDI library.

    Parameters
    ----------
    score : Score
        The Score object to export.
    filename : str
        The path to the output MIDI file.
    format : str
        The export format, should be "midi" for this function.
    show : bool
        Print a text representation of the data.
    display : bool
        Display the MIDI file (always False for MIDI export, included for
        consistency with other export functions).
    """
    global tied_to_notes  # helps to merge tied notes
    tied_to_notes = {}

    score.convert_to_seconds()
    score.merge_tied_notes()

    # 600 gives 1 ms resolution at 100 bpm
    pmscore = pm.PrettyMIDI(resolution=600)

    # Create tempo changes from TimeMap
    tm = score.time_map
    if tm is not None:
        tick_scales = []
        for i in range(len(tm.changes)):
            bpm = tm.get_tempo_at(i)
            resolution = pmscore.resolution
            # form breakpoints with units of (ticks, seconds/tick)
            tick_scale = 60.0 / (bpm * resolution)
            tick_scales.append(
                (int(tm.changes[i].quarter * resolution), tick_scale)
            )
        pmscore._tick_scales = tick_scales

    # We want to write every Part or Staff as a separate PrettyMIDI Instrument
    # Gather the EventGroups to write. Note that the reader may reconstruct
    # The Part/Staff hierarchy if all Parts have different instruments and
    # read_score is invoked with the default group_by_instrument=True.
    evgroups: list[Part | Staff] = []
    for part in score.find_all(Part):  # type: ignore
        # search Part for any Notes at the top level
        part = cast(Part, part)
        if any(isinstance(ev, Note) for ev in part.content):
            evgroups.append(part)
        # search for Staffs within Part
        for staff in part.find_all(Staff):
            staff = cast(Staff, staff)
            evgroups.append(staff)
    key_signatures = []  # Collect key signatures for pmscore

    # Create instruments and add notes
    for evgroup in evgroups:
        # determine a name for the pm.Instrument
        if isinstance(evgroup, Part):
            part: Part = evgroup
        else:
            part = evgroup.part  # type: ignore
        name = part.instrument
        program = part.get("midi_program")
        if name is not None and program is None:
            try:
                program = pm.instrument_name_to_program(name)
            except Exception:
                program = None
        if program is None:
            program = 0  # Acoustic Grand Piano as default
        instrument = pm.Instrument(
            program, name=name if name is not None else "Unknown"
        )

        add_eventgroup_to_instrument(evgroup, instrument, key_signatures)
        pmscore.instruments.append(instrument)

    # Create time signature changes
    for ts in score.time_signatures:
        # Convert onset (in quarters) to seconds using time_map
        time_in_seconds = score.time_map.quarter_to_time(ts.time)
        pm_ts = pm.TimeSignature(
            numerator=int(ts.upper),
            denominator=int(ts.lower),
            time=time_in_seconds,
        )
        pmscore.time_signature_changes.append(pm_ts)

    # Add key signatures to pmscore
    pmscore.key_signature_changes = key_signatures

    # Force pretty_midi to use custom _tick_scales before writing:
    _ = pmscore.get_end_time()

    if show:
        from amads.io.pm_show import pretty_midi_show

        pretty_midi_show(pmscore, filename)

    # Write to MIDI file
    pmscore.write(filename)
    tied_to_notes = None  # clear global variable after use



from typing import Iterable

from pytest import approx

from amads.algorithms.slice.window import sliding_window
from amads.core.basics import Note, Score


def test_sliding_window(twochan_score: Score, twochan_notes: Iterable[Note]):
    size = 4.0  # one bar
    step = 0.01

    score = twochan_score.flatten(collapse=True)

    last_note = twochan_notes[-1]
    last_note_off = last_note.offset
    assert last_note_off == approx(16.0)

    windows = sliding_window(
        score,
        size=size,
        step=step,
        align="right",
    )
    windows = list(windows)

    first_slice = windows[0]
    assert first_slice.onset == approx(-4.0)
    assert first_slice.offset == approx(0.0)
    assert len(first_slice.content) == 0

    last_slice = windows[-1]
    assert last_slice.onset == approx(last_note_off, abs=step * 3)
    assert len(last_slice.content) == 0

    # The second slice should include the score's opening two notes
    assert len(windows[1].content) == 2
    assert windows[1].content[0].key_num == 67 - 2 * 12
    assert windows[1].content[1].key_num == 67
